//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "BitmapProcesorUnit.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------

//   Important: Methods and properties of objects in VCL can only be
//   used in a method called using Synchronize, for example:
//
//      Synchronize(UpdateCaption);
//
//   where UpdateCaption could look like:
//
//      void __fastcall Unit1::UpdateCaption()
//      {
//        Form1->Caption = "Updated in a thread";
//      }
//---------------------------------------------------------------------------
__fastcall TBitmapProcesorThread::TBitmapProcesorThread(bool CreateSuspended)
        : TThread(CreateSuspended)
{
//  Access=new TCriticalSection();
}
//---------------------------------------------------------------------------
__fastcall TBitmapProcesorThread::~TBitmapProcesorThread()
{
/*  if (Access!=NULL)
    Access->Acquire();
  TCriticalSection *oAccess=Access;
  Access=NULL;
    oAccess->Release();
  delete oAccess;*/
}
//---------------------------------------------------------------------------
void __fastcall TBitmapProcesorThread::Execute()
{
    switch (function)
      {
      case tfNothing:
          {
          };break;
      case tfResizeBitmap:
          {
          ResizeBitmap(destPict,srcPict,fmode);
          };break;
      case tfFlipBitmap:
          {
          FlipBitmap(destPict,srcPict,(TScrollBarKind)extParamInt);
          };break;
      case tfCreateHistBitmap:
          {
          CountHistogram(htNormalized,destBufR,destBufG,destBufB,srcPict);
          };break;
      case tfCurvLevelsBitmap:
          {
          ConductLevelsChange(destPict,curveParam,(destChannel)extParamInt);
          };break;
      case tfAutoLevelsBitmap:
          {
          ConductAutoLevels(destPict,(destChannel)extParamInt);
          };break;
      case tfGaussianFilter:
          {
          GaussianBlurBitmap(destPict,srcPict,extParam);
          };break;
      case tfTriangularFilter:
          {
          TriangularBlurBitmap(destPict,srcPict,extParam);
          };break;
      case tfSharpenFilter:
          {
          SharpenBitmap(destPict,srcPict,extParam,extParamAng);
          };break;
      case tfEdgesFilter:
          {
          FindEdgesBitmap(destPict,srcPict,extParam,extParamAng);
          };break;
      case tfExternalFilter:
          {
          FloatFilterBitmap(destPict,srcPict,FilterBuf,extParamInt,extParam,extParamAng,extParamFlag);
          };break;
      case tfGaussianPaint:
          {
          MakeMonochromeBitmapPalette(destPict);
          FillBitmapWithGauss2D(destPict,extParam,newWidthParam,newHeightParam);
          };break;
      case tfTriangularPaint:
          {
          MakeMonochromeBitmapPalette(destPict);
          FillBitmapWithRotatedTriangle2D(destPict,extParam,extParamAng,newWidthParam,newHeightParam);
          };break;
      case tfAbstractPaint:
          {
          MakeMonochromeBitmapPalette(destPict);
          FillBitmapWithAbstract2D(destPict,extParamInt,extParam,newWidthParam,newHeightParam);
          };break;
      case tfSharpenPaint:
          {
          //MakeMonochromeBitmapPalette(destPict);
          MakeSignedRGBitmapPalette(destPict);
          FillBitmapWithSharpenFilter2D(destPict,extParam,newWidthParam,newHeightParam);
          };break;
      case tfEdgesPaint:
          {
          MakeSignedRGBitmapPalette(destPict);
          FillBitmapWithEdgeFilter2D(destPict,extParam,newWidthParam,newHeightParam);
          };break;
      case tfFFTBitmap:
          {
          ConductFFTBitmap(destPict,srcPict);
          };break;
      case tfRFTBitmap:
          {
          ConductRFTBitmap(destPict,srcPict);
          };break;
      case tfFFTGaussFilter:
          {
          FFTGaussianBlurBitmap(destPict,srcPict,extParam);
          };break;
      }

}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::MakeMonochromeBitmapPalette(Graphics::TBitmap *dest)
{
  dest->PixelFormat=pf8bit;
  TLogPalEx lPalEx;
  lPalEx.lPal.palVersion = 0x300;
  lPalEx.lPal.palNumEntries = 256;
  for (int i=0;i<lPalEx.lPal.palNumEntries;i++)
    {
    lPalEx.lPal.palPalEntry[i].peRed=i;
    lPalEx.lPal.palPalEntry[i].peGreen=i;
    lPalEx.lPal.palPalEntry[i].peBlue=i;
    lPalEx.lPal.palPalEntry[i].peFlags=0;
    }
  dest->Palette = CreatePalette(&lPalEx.lPal);
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::MakeSignedRGBitmapPalette(Graphics::TBitmap *dest)
{
  dest->PixelFormat=pf8bit;
  TLogPalEx lPalEx;
  lPalEx.lPal.palVersion = 0x300;
  lPalEx.lPal.palNumEntries = 256;
  for (int i=0;i<lPalEx.lPal.palNumEntries;i++)
    {
    lPalEx.lPal.palPalEntry[i].peRed=max(i-128,0)*2;
    lPalEx.lPal.palPalEntry[i].peGreen=max(128-i,0)*2;
    lPalEx.lPal.palPalEntry[i].peBlue=max((i-128)/3,(128-i)/2);
    lPalEx.lPal.palPalEntry[i].peFlags=0;
    }
  dest->Palette = CreatePalette(&lPalEx.lPal);
}
//---------------------------------------------------------------------------
/* WERSJA PENA
void TBitmapProcesorThread::ResizeBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,eFilter fmode)
{
  if ((src==NULL)||(dest==NULL)) return;
  int sWidth=src->Width;
  int sHeight=src->Height;
  dWidth=dest->Width;
  dHeight=dest->Height;
  if ((sWidth<1)||(sHeight<1)||(dWidth<1)||(dHeight<1)) return;
  //Tworzymy tablic filtru
  Graphics::TBitmap *filter=new Graphics::TBitmap();
  MakeMonochromeBitmapPalette(filter);

  switch (fmode)
    {
    case kPoint:
      {
      filter->Width=1;
      filter->Height=1;
      ((uchar *)filter->ScanLine[0])[0]=1;
      };break;
    case kLinearInterp:
      {
      Extended ffWidth;
      Extended ffHeight;
      ffWidth=max(Extended(sWidth)/dWidth,Extended(dWidth)/sWidth)*(Extended(1)+1e-15);
      ffHeight=max(Extended(sHeight)/dHeight,Extended(dHeight)/sHeight)*(Extended(1)+1e-15);
      FillBitmapWithTriangle2D(filter,stromoscT,ffWidth,ffHeight);
      };break;
    case kCubicInterp:
      {
      Extended ffWidth;
      Extended ffHeight;
      //Ustalamy wymiary - zalenie czy zwiekszamy czy zmniejszamy
      if (sWidth>=dWidth)
          {//Pomniejszamy szerok
          ffWidth=Extended(sWidth)/dWidth;
          }
        else
          {//Powikszamy szeroko
          ffWidth=Extended(dWidth)/sWidth;
          if ((ffWidth>1)) ffWidth+=1;
          }
      if (sHeight>=dHeight)
          {//Pomniejszamy wysoko
          ffHeight=Extended(sHeight)/dHeight;
          }
        else
          {//Powikszamy wysoko
          ffHeight=Extended(dHeight)/sHeight;
          if ((ffHeight>1)) ffHeight+=1;
          }
      FillBitmapWithGauss2D(filter,stromoscG,ffWidth,ffHeight);
      }
    }
  //Tablica gotowa, pora na filtracj:
  ResizeWithConvolution(dest,src,filter);
  //dest->Assign(filter);
  delete filter;
}
*/
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ResizeBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,eFilter fmode)
//Wersja uproszczona dla wikszej szybkoci
{
  if ((src==NULL)||(dest==NULL)) return;
  int sWidth=src->Width;
  int sHeight=src->Height;
  dWidth=dest->Width;
  dHeight=dest->Height;
  if ((sWidth<1)||(sHeight<1)||(dWidth<1)||(dHeight<1)) return;
  //Tworzymy tablic filtru
  Graphics::TBitmap *filter=new Graphics::TBitmap();
  MakeMonochromeBitmapPalette(filter);
      {
      Extended ffWidth;
      Extended ffHeight;
      //Ustalamy wymiary - zalenie czy zwiekszamy czy zmniejszamy
      if (sWidth>=dWidth)
          {//Pomniejszamy szerok
          ffWidth=Extended(sWidth)/dWidth;
          }
        else
          {//Powikszamy szeroko
          ffWidth=Extended(dWidth)/sWidth;
          if ((ffWidth>1)) ffWidth+=1;
          }
      if (sHeight>=dHeight)
          {//Pomniejszamy wysoko
          ffHeight=Extended(sHeight)/dHeight;
          }
        else
          {//Powikszamy wysoko
          ffHeight=Extended(dHeight)/sHeight;
          if ((ffHeight>1)) ffHeight+=1;
          }
      FillBitmapWithGauss2D(filter,stromoscG,ffWidth,ffHeight);
      }
  //Tablica gotowa, pora na filtracj:
  ResizeWithConvolution(dest,src,filter);
  delete filter;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::GaussianBlurBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended level)
{
  if ((dest==NULL)||(src==NULL)) return;
  dWidth=src->Width;
  dHeight=src->Height;
  if ((dWidth<1)||(dHeight<1)) return;
  //Tworzymy tablic filtru
  Graphics::TBitmap *filter=new Graphics::TBitmap();
  MakeMonochromeBitmapPalette(filter);
  Extended ffSize;
  ffSize=level+1;
  FillBitmapWithGauss2D(filter,stromoscG,ffSize,ffSize);
  //Tablica gotowa, pora na filtracj:
  ConductConvolutionUnsgn(dest,src,filter);
  //dest->Assign(filter);
  delete filter;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FloatFilterBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended *filter,int fSize,Extended fDivider,Extended fAdder,bool UseAbsVal)
{
  if ((dest==NULL)||(src==NULL)) return;
  dWidth=src->Width;
  dHeight=src->Height;
  if ((dWidth<1)||(dHeight<1)) return;
  //Filtrujemy:
  ConductConvolutionSign(dest,src,filter,fSize,fDivider,fAdder,UseAbsVal);
  delete filter;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::TriangularBlurBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended level)
{
  if ((dest==NULL)||(src==NULL)) return;
  dWidth=src->Width;
  dHeight=src->Height;
  if ((dWidth<1)||(dHeight<1)) return;
  //Tworzymy tablic filtru
  Graphics::TBitmap *filter=new Graphics::TBitmap();
  MakeMonochromeBitmapPalette(filter);
  Extended ffSize=level+1;
  FillBitmapWithTriangle2D(filter,stromoscT,ffSize,ffSize);
  //Tablica gotowa, pora na filtracj:
  ConductConvolutionUnsgn(dest,src,filter);
  //dest->Assign(filter);
  delete filter;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::SharpenBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended level,Extended size)
{
  if ((dest==NULL)||(src==NULL)) return;
  dWidth=src->Width;
  dHeight=src->Height;
  if ((dWidth<1)||(dHeight<1)) return;
  //Tworzymy tablic filtru
  Graphics::TBitmap *filter=new Graphics::TBitmap();
  MakeSignedRGBitmapPalette(filter);
  Extended ffSize=size+1;
  FillBitmapWithSharpenFilter2D(filter,level*stromoscG,ffSize,ffSize);
  //Tablica gotowa, pora na filtracj:
  ConductConvolutionSign(dest,src,filter);
  //dest->Assign(filter);
  delete filter;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FindEdgesBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended level,Extended size)
{
  if ((dest==NULL)||(src==NULL)) return;
  dWidth=src->Width;
  dHeight=src->Height;
  if ((dWidth<1)||(dHeight<1)) return;
  //Tworzymy tablic filtru
  Graphics::TBitmap *filter=new Graphics::TBitmap();
  MakeSignedRGBitmapPalette(filter);
  Extended ffSize=size;
  FillBitmapWithEdgeFilter2D(filter,level,ffSize,ffSize);
  //Tablica gotowa, pora na filtracj:
  ConductConvolutionSign(dest,src,filter);
  //dest->Assign(filter);
  delete filter;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ResizeWithConvolution(Graphics::TBitmap *dest,Graphics::TBitmap *src,Graphics::TBitmap *filter)
{
  int sWidth=src->Width;
  int sHeight=src->Height;
  dWidth=dest->Width;
  dHeight=dest->Height;
  if ((sWidth<1)||(sHeight<1)||(dWidth<1)||(dHeight<1)) return;
  //Aktualizujemy pasek stanu
  AdvanceMax=Extended(dWidth)*dHeight;
  i=0;j=0;
  Synchronize(UpdateAdvancePos);
  //Dopasowujemy sobie parametry obrazkw
  src->PixelFormat=pf24bit;
  dest->PixelFormat=pf24bit;
  //Liczymy sum wartoci filtra (dzielnik jasnoci)
  int fWidth=filter->Width;
  int fHeight=filter->Height;
  Extended fSum=0.001;
  Extended Scale;
    {
    for (int fi=0;fi<fHeight;fi++)
      for (int fj=0;fj<fWidth;fj++)
        fSum+=((uchar *)filter->ScanLine[fi])[fj];
    Scale=1.001/fSum;
    }
  Extended fWidthDiv2=Extended(fWidth)/2;
  Extended fHeightDiv2=Extended(fHeight)/2;
  //Przemiatanie po obrazku docelowym
  for (i=0;i<dHeight;i++)
    {
    //Aktualizujemy pasek stanu
    if (Terminated) return;
    Synchronize(UpdateAdvancePos);
    for (j=0;j<dWidth;j++)
        {
        //Liczymy now warto piksela
        Extended ValueR=0;
        Extended ValueG=0;
        Extended ValueB=0;
        Extended srcLineIdx;
        Extended srcPixelIdx;
        srcLineIdx=Extended(sHeight)*Extended(i)/max(dHeight-1,1);
        srcPixelIdx=Extended(sWidth)*Extended(j)/max(dWidth-1,1);
        for (int fi=0;fi<fHeight;fi++)
            {
            Extended lineIdxEx;
            if (sHeight>=dHeight)
                //Pomniejszanie
                lineIdxEx=srcLineIdx+(-fHeightDiv2+fi)+0.5;
              else
                //Zwikszanie
                lineIdxEx=srcLineIdx+(-fHeightDiv2+fi+0.5)*sHeight/dHeight;
            int lineIdx=max(Floor(lineIdxEx),0);
            lineIdx=min(lineIdx,sHeight-1);
            for (int fj=0;fj<fWidth;fj++)
              {
              Extended pixIdxEx;
              if (sWidth>=dWidth)
                  //Pomniejszanie
                  pixIdxEx=srcPixelIdx+(-fWidthDiv2+fj)+0.5;
                else
                  //Zwikszanie
                  pixIdxEx=srcPixelIdx+(-fWidthDiv2+fj+0.5)*sWidth/dWidth;
              int pixIdx=max(Floor(pixIdxEx),0);
              pixIdx=min(pixIdx,sWidth-1)*3;
              ValueR+=((uchar *)filter->ScanLine[fi])[fj] * ((uchar *)src->ScanLine[lineIdx])[pixIdx];
              ValueG+=((uchar *)filter->ScanLine[fi])[fj] * ((uchar *)src->ScanLine[lineIdx])[pixIdx+1];
              ValueB+=((uchar *)filter->ScanLine[fi])[fj] * ((uchar *)src->ScanLine[lineIdx])[pixIdx+2];
              }
            }
        //Wstawiamy warto (pierwszy wymiar jest pionowy)
        ((uchar *)dest->ScanLine[i])[j*3]=Floor(Scale*ValueR);
        ((uchar *)dest->ScanLine[i])[j*3+1]=Floor(Scale*ValueG);
        ((uchar *)dest->ScanLine[i])[j*3+2]=Floor(Scale*ValueB);
        }
    }
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductConvolutionUnsgn(Graphics::TBitmap *dest,Graphics::TBitmap *src,Graphics::TBitmap *filter)
//Jeeli filtr ma wymiar parzysty, element centralny jest zaokrglany w d
{
  //Liczymy rodek filtru
  Extended fWidthCenter=Extended(filter->Width)/2;
  Extended fHeightCenter=Extended(filter->Height)/2;
  //I robimy splot
  ConductConvolutionUnsgn(dest,src,filter,fWidthCenter,fHeightCenter);
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductConvolutionUnsgn(Graphics::TBitmap *dest,Graphics::TBitmap *src,Graphics::TBitmap *filter,Extended fWidthCenter,Extended fHeightCenter)
//Przeprowadza splot wzgldem elementu filtra podanego liczbami extended
{
  dWidth=src->Width;
  dHeight=src->Height;
  if ((dWidth<1)||(dHeight<1)) return;
  //Aktualizujemy pasek stanu
  AdvanceMax=Extended(dWidth)*dHeight;
  i=0;j=0;
  Synchronize(UpdateAdvancePos);
  //Dopasowujemy sobie parametry obrazkw
  src->PixelFormat=pf24bit;
  dest->PixelFormat=pf24bit;
  dest->Width=dWidth;
  dest->Height=dHeight;
  //Liczymy sum wartoci filtra (dzielnik jasnoci)
  int fWidth=filter->Width;
  int fHeight=filter->Height;
  Extended fSum=0.001;
  Extended Scale;
    {
    for (int fi=0;fi<fHeight;fi++)
      for (int fj=0;fj<fWidth;fj++)
        fSum+=((uchar *)filter->ScanLine[fi])[fj];
    Scale=1.001/fSum;
    }
  //Przemiatanie po obrazku docelowym
  for (i=0;i<dHeight;i++)
    {
    //Aktualizujemy pasek stanu
    if (Terminated) return;
    Synchronize(UpdateAdvancePos);
    for (j=0;j<dWidth;j++)
        {
        //Liczymy now warto piksela
        Extended ValueR=0;
        Extended ValueG=0;
        Extended ValueB=0;
        int heightParam=Floor(i-fHeightCenter+0.51);
        int widthParam=Floor(j-fWidthCenter+0.51);
        for (int fi=0;fi<fHeight;fi++)
            {
            int lineIdx=max(fi+heightParam,0);
              lineIdx=min(lineIdx,dHeight-1);
            for (int fj=0;fj<fWidth;fj++)
              {
              int pixIdx=max(fj+widthParam,0);
              pixIdx=min(pixIdx,dWidth-1)*3;
              ValueR+=((uchar *)filter->ScanLine[fi])[fj] * ((uchar *)src->ScanLine[lineIdx])[pixIdx];
              ValueG+=((uchar *)filter->ScanLine[fi])[fj] * ((uchar *)src->ScanLine[lineIdx])[pixIdx+1];
              ValueB+=((uchar *)filter->ScanLine[fi])[fj] * ((uchar *)src->ScanLine[lineIdx])[pixIdx+2];
              }
            }
        //Wstawiamy warto (pierwszy wymiar jest pionowy)
        ((uchar *)dest->ScanLine[i])[j*3]=Floor(Scale*ValueR);
        ((uchar *)dest->ScanLine[i])[j*3+1]=Floor(Scale*ValueG);
        ((uchar *)dest->ScanLine[i])[j*3+2]=Floor(Scale*ValueB);
        }
    }
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductConvolutionSign(Graphics::TBitmap *dest,Graphics::TBitmap *src,Graphics::TBitmap *filter)
//Jeeli filtr ma wymiar parzysty, element centralny jest zaokrglany w d
//Liczby poniej 128 traktowane s jako ujemne
{
  //Liczymy rodek filtru
  Extended fWidthCenter=Extended(filter->Width)/2;
  Extended fHeightCenter=Extended(filter->Height)/2;
  //I robimy splot
  ConductConvolutionSign(dest,src,filter,fWidthCenter,fHeightCenter);
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductConvolutionSign(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended *filter,int fSize,Extended fDivider,Extended fAdder,bool UseAbsVal)
//Jeeli filtr ma wymiar parzysty, element centralny jest zaokrglany w d
//Liczby poniej 128 traktowane s jako ujemne
{
  //Liczymy rodek filtru
  Extended fWidthCenter=Extended(fSize-1)/2;
  Extended fHeightCenter=Extended(fSize-1)/2;
  //I robimy splot
  ConductConvolutionSign(dest,src,filter,fSize,fWidthCenter,fHeightCenter,fDivider,fAdder,UseAbsVal);
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductConvolutionSign(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended *filter,int fSize,Extended fWidthCenter,Extended fHeightCenter,Extended fDivider,Extended fAdder,bool UseAbsVal)
//Jeeli filtr ma wymiar parzysty, element centralny jest zaokrglany w d
//Liczby s normalizowane przez dzielnik, ktry powinien by rwny sumie wspczynnikw.
{
  dWidth=src->Width;
  dHeight=src->Height;
  if ((dWidth<1)||(dHeight<1)) return;
  //Aktualizujemy pasek stanu
  AdvanceMax=Extended(dWidth)*dHeight;
  i=0;j=0;
  Synchronize(UpdateAdvancePos);
  //Dopasowujemy sobie parametry obrazkw
  src->PixelFormat=pf24bit;
  dest->PixelFormat=pf24bit;
  dest->Width=dWidth;
  dest->Height=dHeight;
  //Liczymy sum wartoci filtra (dzielnik jasnoci)
  int fWidth=fSize;
  int fHeight=fSize;
  Extended Scale=1.00001/fDivider;
  //Przemiatanie po obrazku docelowym
  for (i=0;i<dHeight;i++)
    {
    for (j=0;j<dWidth;j++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        Synchronize(UpdateAdvancePos);
        //Liczymy now warto piksela
        Extended ValueR=0;
        Extended ValueG=0;
        Extended ValueB=0;
        //Liczymy wsprzdne, od ktrych mamy zacz nakadanie filtru
        int heightParam=Floor(i-fHeightCenter+0.5);
        int widthParam=Floor(j-fWidthCenter+0.5);
        //I nakadamy filtr na mapk liczc jeden piksel
        for (int fi=0;fi<fHeight;fi++)
            {
            //Liczymy konkretne wsprzdne dla punktu filtru
            int lineIdx=max(fi+heightParam,0);
              lineIdx=min(lineIdx,dHeight-1);
            for (int fj=0;fj<fWidth;fj++)
              {
              int pixIdx=max(fj+widthParam,0);
              pixIdx=min(pixIdx,dWidth-1)*3;
              ValueR+=(filter[fWidth*fi+fj] * ((uchar *)src->ScanLine[lineIdx])[pixIdx]);
              ValueG+=(filter[fWidth*fi+fj] * ((uchar *)src->ScanLine[lineIdx])[pixIdx+1]);
              ValueB+=(filter[fWidth*fi+fj] * ((uchar *)src->ScanLine[lineIdx])[pixIdx+2]);
              }
            }
        //Wstawiamy warto piksela nowej mapy(pierwszy wymiar jest pionowy)
        if (UseAbsVal)
          {
          ValueR=fabsl(Scale*ValueR+fAdder);
          ValueG=fabsl(Scale*ValueG+fAdder);
          ValueB=fabsl(Scale*ValueB+fAdder);
          }
         else
          {
          ValueR=(Scale*ValueR+fAdder);
          ValueG=(Scale*ValueG+fAdder);
          ValueB=(Scale*ValueB+fAdder);
          }
        ValueR=min(ValueR,Extended(255.0));
        ValueG=min(ValueG,Extended(255.0));
        ValueB=min(ValueB,Extended(255.0));
        ValueR=max(ValueR,Extended(0.0));
        ValueG=max(ValueG,Extended(0.0));
        ValueB=max(ValueB,Extended(0.0));
        ((uchar *)dest->ScanLine[i])[j*3]=Floor(ValueR);
        ((uchar *)dest->ScanLine[i])[j*3+1]=Floor(ValueG);
        ((uchar *)dest->ScanLine[i])[j*3+2]=Floor(ValueB);
        }
    }
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductConvolutionSign(Graphics::TBitmap *dest,Graphics::TBitmap *src,Graphics::TBitmap *filter,Extended fWidthCenter,Extended fHeightCenter)
//Jeeli filtr ma wymiar parzysty, element centralny jest zaokrglany w d
//Liczby poniej 128 traktowane s jako ujemne
{
  dWidth=src->Width;
  dHeight=src->Height;
  if ((dWidth<1)||(dHeight<1)) return;
  //Aktualizujemy pasek stanu
  AdvanceMax=Extended(dWidth)*dHeight;
  i=0;j=0;
  Synchronize(UpdateAdvancePos);
  //Dopasowujemy sobie parametry obrazkw
  src->PixelFormat=pf24bit;
  dest->PixelFormat=pf24bit;
  dest->Width=dWidth;
  dest->Height=dHeight;
  //Liczymy sum wartoci filtra (dzielnik jasnoci)
  int fWidth=filter->Width;
  int fHeight=filter->Height;
  Extended fSum=0.001;
  Extended Scale;
    {
    for (int fi=0;fi<fHeight;fi++)
      for (int fj=0;fj<fWidth;fj++)
        {
        fSum+=(((uchar *)filter->ScanLine[fi])[fj]);
        fSum-=128;
        }
    Scale=1.001/fSum;
    }
  //Przemiatanie po obrazku docelowym
  for (i=0;i<dHeight;i++)
    {
    for (j=0;j<dWidth;j++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        Synchronize(UpdateAdvancePos);
        //Liczymy now warto piksela
        Extended ValueR=0;
        Extended ValueG=0;
        Extended ValueB=0;
        int heightParam=Floor(i-fHeightCenter+0.5);
        int widthParam=Floor(j-fWidthCenter+0.5);
        for (int fi=0;fi<fHeight;fi++)
            {
            int lineIdx=max(fi+heightParam,0);
              lineIdx=min(lineIdx,dHeight-1);
            for (int fj=0;fj<fWidth;fj++)
              {
              int pixIdx=max(fj+widthParam,0);
              pixIdx=min(pixIdx,dWidth-1)*3;
              int filterR=(((uchar *)filter->ScanLine[fi])[fj]);
              filterR-=128;
              int filterG=(((uchar *)filter->ScanLine[fi])[fj]);
              filterG-=128;
              int filterB=(((uchar *)filter->ScanLine[fi])[fj]);
              filterB-=128;
              ValueR+=(filterR * ((uchar *)src->ScanLine[lineIdx])[pixIdx]);
              ValueG+=(filterG * ((uchar *)src->ScanLine[lineIdx])[pixIdx+1]);
              ValueB+=(filterB * ((uchar *)src->ScanLine[lineIdx])[pixIdx+2]);
              }
            }
        //Wstawiamy warto (pierwszy wymiar jest pionowy)
        ValueR=min(Scale*ValueR,Extended(255.0));
        ValueG=min(Scale*ValueG,Extended(255.0));
        ValueB=min(Scale*ValueB,Extended(255.0));
        ValueR=max(ValueR,Extended(0.0));
        ValueG=max(ValueG,Extended(0.0));
        ValueB=max(ValueB,Extended(0.0));
        ((uchar *)dest->ScanLine[i])[j*3]=Floor(ValueR);
        ((uchar *)dest->ScanLine[i])[j*3+1]=Floor(ValueG);
        ((uchar *)dest->ScanLine[i])[j*3+2]=Floor(ValueB);
        }
    }
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FillBitmapWithGauss2D(Graphics::TBitmap *filter,Extended stromosc,Extended ffWidth,Extended ffHeight)
//Tworzy w bitmapie filtr Gaussa, o rozmiarach podanych parametrami zmiennoprzecinkowymi.
//Bitmapa musi by 8-bitowa i miec palet z 256-cioma kolorami.
//Element centralny bitmapy jest na rodku (rozm. niepaysty) lub
// ma wsp. zaokrglone w d (rozm. parzysty).
{
      int fWidth=Ceil(ffWidth);  //Ceil zaokrgla w gr
      int fHeight=Ceil(ffHeight);
      filter->Width=fWidth;
      filter->Height=fHeight;
      if (stromosc<=0) stromosc=1e-30;
      //Minimalna stromosc dajca rozsdny wynik to okoo 0.00196
      Extended StromFactH2=logl(stromosc)/(ffHeight*ffHeight);
      Extended StromFactW2=logl(stromosc)/(ffWidth*ffWidth);
      for (int i=0;i<fHeight;i++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        //Synchronize(UpdateAdvancePos);
        //I liczymy ...
        Extended AmpFactor=(Extended(i)-(0.5*(ffHeight-1)));//+0.505
        Extended Amplitude=Extended(255.0)*expl(-StromFactH2*IntPower(AmpFactor,2));
        for (int j=0;j<fWidth;j++)
          {
          Extended ValFactor=(Extended(j)-(0.5*(ffWidth-1)));//+0.505
          Extended Value=Amplitude*expl(-StromFactW2*IntPower(ValFactor,2));
          //Pierwszy wymiar jest pionowy
          if (Value>MaxInt)
              Value=fmodl(Value, 256);
          ((uchar *)filter->ScanLine[i])[j]=Floor(Value);
          }
        }
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FillTableWithGauss2D(Extended *filter,uint fWidth,uint fHeight,Extended stromoscH,Extended stromoscW,Extended ffCenterW,Extended ffCenterH)
//Tworzy w tablicy Extended filtr Gaussa, o rozmiarach podanych parametrami zmiennoprzecinkowymi.
//Element centralny filtru jest na rodku (rozm. niepaysty) lub
// ma wsp. zaokrglone w d (rozm. parzysty).
{
      for (uint i=0;i<fHeight;i++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        //Synchronize(UpdateAdvancePos);
        //I liczymy ...
        Extended AmpFactor=(Extended(i)-ffCenterH);//(0.5*(ffHeight-1))
        Extended Amplitude=expl(-stromoscH*AmpFactor*AmpFactor);
        for (uint j=0;j<fWidth;j++)
          {
          Extended ValFactor=(Extended(j)-ffCenterW);//(0.5*(ffWidth-1))
          //Pierwszy wymiar jest pionowy
          filter[i*fWidth+j]=Amplitude*expl(-stromoscW*ValFactor*ValFactor);
          }
        }

      Extended sum=0.0;
      for (uint i=0;i<fHeight;i++)
        for (uint j=0;j<fWidth;j++)
          sum+=filter[i*fWidth+j];
      if (((sum<0.99999)||(sum>1.00001))&&((sum>0.0001)||(sum<-0.0001)))
        {
        Extended Multiplier;
        if (sum>0)
            Multiplier=1/sum;
          else
            Multiplier=-1/sum;
        for (uint i=0;i<fHeight;i++)
          for (uint j=0;j<fWidth;j++)
            filter[i*fWidth+j]*=Multiplier;
        }

}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FillBitmapWithSharpenFilter2D(Graphics::TBitmap *filter,Extended stromosc,Extended ffWidth,Extended ffHeight)
//Tworzy w bitmapie filtr wyostrzajcy, o rozmiarach podanych parametrami zmiennoprzecinkowymi.
//Bitmapa musi by 8-bitowa i miec palet z 256-cioma kolorami.
//Element centralny bitmapy jest na rodko (rozm. niepaysty) lub
// ma wsp. zaokrglone w d (rozm. parzysty).
//Aby uzyska wartoci z przedziau (-128,127) naley odj od wartoci filtru 128
//(tak robi funkcja do splotu ze znakiem)
{
      //Najpierw liczymy Gaussa
      FillBitmapWithGauss2D(filter,stromosc,ffWidth,ffHeight);
      //Potem na gaussa nakadamy sinusa
        //Najpierw parametry sinusa - koniec mapy musi byc na 2*pi
      int fWidth=Ceil(ffWidth);  //Ceil zaokrgla w gr
      int fHeight=Ceil(ffHeight);
      Extended ScaleFactor=logl(stromosc/stromoscG);
      Extended WidthScale=(2.0*M_PI)/ffWidth*ScaleFactor;
      Extended HeightScale=(2.0*M_PI)/ffHeight*ScaleFactor;
      for (int i=0;i<fHeight;i++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        //Synchronize(UpdateAdvancePos);
        //I liczymy ...
        for (int j=0;j<fWidth;j++)
          {
          Extended length=sqrtl(IntPower(HeightScale*(Extended(i)+0.5-(ffHeight/2)),2)+IntPower(WidthScale*(Extended(j)+0.5-(ffWidth/2)),2));
          if (length>(M_PI)) length=(M_PI);
          Extended Value=((uchar *)filter->ScanLine[i])[j];
          Value*=cosl(length)/2;
          //Pierwszy wymiar jest pionowy
          if (Value>MaxInt)
              Value=fmodl(Value, 128);
          Value+=128;
          if (Value<0) Value=0;
          ((uchar *)filter->ScanLine[i])[j]=Floor(Value);
          }
        }
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FillBitmapWithEdgeFilter2D(Graphics::TBitmap *filter,Extended level,Extended ffWidth,Extended ffHeight)
//Tworzy w bitmapie filtr szukajcy krawdzi, o rozmiarach podanych parametrami zmiennoprzecinkowymi.
//Bitmapa musi by 8-bitowa i miec palet z 256-cioma kolorami.
//Element centralny bitmapy jest na rodko (rozm. niepaysty) lub
// ma wsp. zaokrglone w d (rozm. parzysty).
//Aby uzyska wartoci z przedziau (-128,127) naley odj od wartoci filtru 128
//(tak robi funkcja do splotu ze znakiem)
{
        //Najpierw parametry sinusa - koniec mapy musi byc na 2*pi
      int fWidth=Ceil(ffWidth);  //Ceil zaokrgla w gr
      int fHeight=Ceil(ffHeight);
      //Zmieniamy rozmiar bitmapy
      filter->Width=fWidth;
      filter->Height=fHeight;

      Extended WidthScale=(2.0*M_PI)/ffWidth;
      Extended HeightScale=(2.0*M_PI)/ffHeight;
      for (int i=0;i<fHeight;i++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        //Synchronize(UpdateAdvancePos);
        //I liczymy ...
        for (int j=0;j<fWidth;j++)
          {
          //Liczymy odleglo od punktu (i=0,j=0):
          Extended hDistance=HeightScale*(Extended(i)+0.5);
          Extended wDistance=WidthScale*(Extended(j)+0.5);
          Extended length=sqrtl(IntPower(hDistance,2)+IntPower(wDistance,2));
          Extended Value=126.8;
          //warunki "za okresem"
          if (length>M_PI*2.0)
                length=M_PI*2.0;
              else
          //Mnozymy listki ujemne razy poziom
          if (length>M_PI)
                Value*=level;
          //warunki "blisko do cianki"
          if (hDistance<(M_PI_2))
                Value*=sinl(hDistance);
          if (wDistance<(M_PI_2))
                Value*=sinl(wDistance);
          //No i normalna funkcja - najwaniejszy sinus
          if (length>M_PI_2)
              Value*=sinl(length);
          //Pierwszy wymiar jest pionowy
          if ((Value>127))
              Value=127;
          Value+=128;
          if (Value<0) Value=0;
          (((uchar *)filter->ScanLine[i])[j])=Floor(Value);
          }
        }
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FillBitmapWithTriangle2D(Graphics::TBitmap *filter,Extended stromosc,Extended ffWidth,Extended ffHeight)
{
      filter->Width=Ceil(ffWidth);  //Ceil zaokrgla w gr
      filter->Height=Ceil(ffHeight);

      Extended WspKierWzrostH=255.0/((ffHeight+1.0)*(0.5-stromosc));
      Extended WspKierOpadanH=255.0/((ffHeight+1.0)*(0.5+stromosc-1));
      Extended WspPrzesunWzrostH=stromosc*(ffHeight+1.0);
      Extended WspPrzesunOpadanH=(ffHeight+1.0)*(1.0-stromosc);

      Extended WspKierWzrostW=1.0/((ffWidth+1.0)*(0.5-stromosc));
      Extended WspKierOpadanW=1.0/((ffWidth+1.0)*(0.5+stromosc-1));
      Extended WspPrzesunWzrostW=stromosc*(ffWidth+1.0);
      Extended WspPrzesunOpadanW=(ffWidth+1.0)*(1.0-stromosc);

      for (int i=0;i<filter->Height;i++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        //Synchronize(UpdateAdvancePos);
        //I liczymy ...
        Extended Amplitude;
        if (1.0+i <= (ffHeight+1.0)*0.5)
            Amplitude=(WspKierWzrostH*(1.0+i-WspPrzesunWzrostH));
          else
            Amplitude=(WspKierOpadanH*(1.0+i-WspPrzesunOpadanH));
        Amplitude=max(Amplitude,Extended(0));
        for (int j=0;j<filter->Width;j++)
          {
          Extended Value;
          if (1.0+j <= (ffWidth+1.0)*0.5)
              Value=(Amplitude*WspKierWzrostW*(1.0+j-WspPrzesunWzrostW));
            else
              Value=(Amplitude*WspKierOpadanW*(1.0+j-WspPrzesunOpadanW));
          //Pierwszy wymiar jest pionowy
          Value=max(Value,Extended(0));
          if (Value>MaxInt)
              Value=fmodl(Value, 256);
          ((uchar *)filter->ScanLine[i])[j]=min(Floor(Value),255);
          }
        }

}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FillBitmapWithRotatedTriangle2D(Graphics::TBitmap *filter,Extended stromosc,Extended angle,Extended ffWidth,Extended ffHeight)
{
      Extended fRotHeight;
      Extended fRotWidth;
      fRotWidth=fRotHeight=sqrtl(IntPower(ffHeight,2)+IntPower(ffWidth,2))*0.66;
      fRotWidth=Ceil(fRotWidth);
      fRotHeight=Ceil(fRotHeight);

      filter->Width=Ceil(fRotWidth);  //Ceil zaokrgla w gr
      filter->Height=Ceil(fRotHeight);

      Extended WspKierWzrostH=255.0/((ffHeight)*(0.5-stromosc));
      Extended WspKierOpadanH=255.0/((ffHeight)*(0.5+stromosc-1));
      Extended WspPrzesunWzrostH=stromosc*(ffHeight);
      Extended WspPrzesunOpadanH=(ffHeight)*(1.0-stromosc);

      Extended WspKierWzrostW=1.0/((ffWidth)*(0.5-stromosc));
      Extended WspKierOpadanW=1.0/((ffWidth)*(0.5+stromosc-1));
      Extended WspPrzesunWzrostW=stromosc*(ffWidth);
      Extended WspPrzesunOpadanW=(ffWidth)*(1.0-stromosc);

      for (i=0;i<filter->Height;i++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        //Synchronize(UpdateAdvancePos);
        //I liczymy ...
        for (j=0;j<filter->Width;j++)
          {
          //Liczymy nowe wsprzdne
          Extended wRadius=sqrtl(IntPower(i+0.5-fRotHeight*0.5,2)+IntPower(j+0.5-fRotWidth*0.5,2));
          Extended tmp=i+0.5-fRotWidth*0.5;
          if (tmp==0) tmp=1e-20;
          Extended wAngle=atan2l(j+0.5-fRotHeight*0.5,tmp)+angle;//np. M_PI_4
          Extended iRot=wRadius*cosl(wAngle)+ffHeight*0.5;
          Extended jRot=wRadius*sinl(wAngle)+ffWidth*0.5;
          //Liczymy warto (w obu wymiarach)
          Extended Amplitude;
          if (iRot <= ffHeight*0.5)
              Amplitude=(WspKierWzrostH*(iRot-WspPrzesunWzrostH));
            else
              Amplitude=(WspKierOpadanH*(iRot-WspPrzesunOpadanH));
          Amplitude=max(Amplitude,Extended(0));
          Extended Value;
          if (jRot <= (ffWidth)*0.5)
              Value=(Amplitude*WspKierWzrostW*(jRot-WspPrzesunWzrostW));
            else
              Value=(Amplitude*WspKierOpadanW*(jRot-WspPrzesunOpadanW));
          //Pierwszy wymiar jest pionowy
          Value=max(Value,Extended(0));
          if (Value>MaxInt)
              Value=fmodl(Value, 256);
          ((uchar *)filter->ScanLine[i])[j]=min(Floor(Value),255);
          }
        }
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FillBitmapWithAbstract2D(Graphics::TBitmap *filter,int Number,Extended level,Extended ffWidth,Extended ffHeight)
{
  //Zaczynamy - zalenie ktry algorytm wybrano...
  switch (Number)
   {
   case 0:
      {
      //I robimy...
      Extended stromosc=200*level;
      filter->Width=Ceil(ffWidth);  //Ceil zaokrgla w gr
      filter->Height=Ceil(ffHeight);
      Extended StromFactH=(stromosc)/(ffHeight);
      Extended StromFactW=(stromosc)/(ffWidth);
      Extended BfactorH=StromFactH*ffHeight*0.5-255;
      Extended BfactorW=StromFactW*ffWidth*0.5-255;
      for (int i=0;i<filter->Height;i++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        //Synchronize(UpdateAdvancePos);
        //I liczymy ...
        Extended Amplitude=(StromFactH*i-BfactorH);
        if (Amplitude<0) Amplitude=-Amplitude;
        for (int j=0;j<filter->Width;j++)
          {
          Extended Value=Amplitude*(StromFactW*j-BfactorW);
          //Pierwszy wymiar jest pionowy
          if (Value<0) Value*=-1;
          ((uchar *)filter->ScanLine[i])[j]=max(Floor(Value),255);
          }
        }
      };return;
   case 1:
      {
      FillBitmapWithGauss2D(filter,stromoscG*expl(-100*level),ffWidth,ffHeight);
      };return;
   case 2:
      {
      //Najpierw liczymy Gaussa
      FillBitmapWithGauss2D(filter,stromoscG*expl(level),ffWidth,ffHeight);
      //Potem na gaussa nakadamy sinusa
        //Najpierw parametry sinusa
      int fWidth=Ceil(ffWidth);  //Ceil zaokrgla w gr
      int fHeight=Ceil(ffHeight);
      Extended WidthScale=ffWidth/(4*M_PI);
      Extended HeightScale=ffHeight/(4*M_PI);
      for (int i=0;i<fHeight;i++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        //Synchronize(UpdateAdvancePos);
        //I liczymy ...
        Extended Amplitude=sinl(HeightScale*i);
        for (int j=0;j<fWidth;j++)
          {
          Extended Value=((uchar *)filter->ScanLine[i])[j];
          Value*=Amplitude;
          Value*=sinl(WidthScale*j);
          //Pierwszy wymiar jest pionowy
          if (Value>MaxInt)
              Value=fmodl(Value, 256);
          Value+=128;
          if (Value<0) Value=0;
          ((uchar *)filter->ScanLine[i])[j]=Floor(Value);
          }
        }
      };return;
   case 3:
      {
      };return;
   case 4:
      {
      };return;
   case 5:
      {
      };return;
   }
}
//---------------------------------------------------------------------------
/* This function solves for x in the equation "x is y% of z". */
Extended SolveForX(Extended Y,Extended Z)
{
//  return floor( Z * (Y * 0.01) );
  return Z * (Y * 0.01);
}

/* This function solves for y in the equation "x is y% of z". */
Extended SolveForY(Extended X,Extended Z)
{
  if (Z==0)
    return 0;
  else
//    return floor( (X * 100.0) / Z );
    return (X * 100.0) / Z;
}
//---------------------------------------------------------------------------
void __fastcall TBitmapProcesorThread::PaintBarOnBitmap(Graphics::TBitmap* AnImage,
    const TRect& PaintRect,Graphics::TColor ForeColor,Graphics::TColor BackColor,TScrollBarKind Kind,Extended PercentDone)
{
  Longint  FillSize;
  Integer  W, H;

  W = PaintRect.Right - PaintRect.Left + 1;
  H = PaintRect.Bottom - PaintRect.Top + 1;

  AnImage->Canvas->Brush->Color = BackColor;
  AnImage->Canvas->FillRect(PaintRect);
  AnImage->Canvas->Pen->Color = ForeColor;
  AnImage->Canvas->Pen->Width = 1;
  AnImage->Canvas->Brush->Color = ForeColor;
  switch(Kind){
    case sbHorizontal:
      FillSize = SolveForX(PercentDone, W);
      if (FillSize > W)
         FillSize = W;
      if (FillSize > 0)
         AnImage->Canvas->FillRect(Rect(PaintRect.Left,
                                        PaintRect.Top,
                                        FillSize,
                                        H));
     break;
   case sbVertical:
     FillSize = SolveForX(PercentDone, H);
     if (FillSize >= H)
       FillSize = H - 1;
       AnImage->Canvas->FillRect(Rect(PaintRect.Left,
                                      H - FillSize,
                                      W,
                                      H));
     break;
  }
}
//---------------------------------------------------------------------------
void __fastcall TBitmapProcesorThread::PaintTextOnBitmap(Graphics::TBitmap* AnImage, const TRect& PaintRect,const String &Str)
{
  Integer X, Y;
  TBltBitmap *OverRect=new TBltBitmap();

  OverRect->MakeLike(AnImage);
  PaintBitmapBackground(OverRect,cmBlackness);
  OverRect->Canvas->Brush->Style = bsClear;
  OverRect->Canvas->Font->Color = clWhite;
  X = (PaintRect.Right - PaintRect.Left + 1 - OverRect->Canvas->TextWidth(Str)) / 2;
  Y = (PaintRect.Bottom - PaintRect.Top + 1 - OverRect->Canvas->TextHeight(Str)) / 2;
  OverRect->Canvas->TextRect(PaintRect, X, Y, Str);
  AnImage->Canvas->CopyMode = cmSrcInvert;
  AnImage->Canvas->Draw(0, 0, OverRect);
}
//---------------------------------------------------------------------------
void __fastcall TBitmapProcesorThread::PaintBitmapBackground(Graphics::TBitmap* AnImage,int CopyMode)
{
  TRect  ARect;
  AnImage->Canvas->CopyMode = CopyMode;
  ARect = Rect(0, 0, AnImage->Width, AnImage->Height);
  AnImage->Canvas->CopyRect(ARect, AnImage->Canvas, ARect);
  AnImage->Canvas->CopyMode = cmSrcCopy;
}
//---------------------------------------------------------------------------
void __fastcall TBitmapProcesorThread::UpdateAdvancePos()
{
        AdvancePos=Extended(i)*dWidth+j;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FlipBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,TScrollBarKind FlipAxis)
{
  //Pobieramy parametry
  if ((dest==NULL)||(src==NULL)) return;
  dWidth=src->Width;
  dHeight=src->Height;
  if ((dWidth<1)||(dHeight<1)) return;
  //Aktualizujemy pasek stanu
  AdvanceMax=Extended(dWidth)*dHeight;
  i=0;j=0;
  Synchronize(UpdateAdvancePos);
  //Dopasowujemy sobie parametry obrazkw
  src->PixelFormat=pf24bit;
  dest->PixelFormat=pf24bit;
  dest->Width=dWidth;
  dest->Height=dHeight;
  //Przemiatanie po obrazku docelowym
  switch (FlipAxis)
    {
    case sbHorizontal:
      {
      for (i=0;i<dHeight;i++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        Synchronize(UpdateAdvancePos);
        for (j=0;j<dWidth;j++)
            {
            //Birzemy warto piksela i zapisujemy w odwrotnym miejscu
            //obrazka rdowego (pierwszy wymiar jest pionowy)
            uint StareJ=dWidth-j-1;
            ((uchar *)dest->ScanLine[i])[j*3]=((uchar *)src->ScanLine[i])[StareJ*3];
            ((uchar *)dest->ScanLine[i])[j*3+1]=((uchar *)src->ScanLine[i])[StareJ*3+1];
            ((uchar *)dest->ScanLine[i])[j*3+2]=((uchar *)src->ScanLine[i])[StareJ*3+2];
            }
        }
      };break;
    case sbVertical:
      {
      for (i=0;i<dHeight;i++)
        {
        //Aktualizujemy pasek stanu
        if (Terminated) return;
        Synchronize(UpdateAdvancePos);
        for (j=0;j<dWidth;j++)
            {
            //Birzemy warto piksela i zapisujemy w odwrotnym miejscu
            //obrazka rdowego (pierwszy wymiar jest pionowy)
            uint StareI=dHeight-i-1;
            ((uchar *)dest->ScanLine[i])[j*3]=((uchar *)src->ScanLine[StareI])[j*3];
            ((uchar *)dest->ScanLine[i])[j*3+1]=((uchar *)src->ScanLine[StareI])[j*3+1];
            ((uchar *)dest->ScanLine[i])[j*3+2]=((uchar *)src->ScanLine[StareI])[j*3+2];
            }
        }
      };break;
    }
}
//---------------------------------------------------------------------------
// C callable FFT function
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FFT(COMPLEX *Y, int N)      //input sample array, # of points
{
 COMPLEX temp1,temp2;             //temporary storage variables
 int i,j,k;                       //loop counter variables
 int upper_leg, lower_leg;   	    //index of upper/lower butterfly leg
 int leg_diff;                    //difference between upper/lower leg
 int num_stages = 0;              //number of FFT stages (iterations)
 int index, step;                 //index/step through twiddle constant
 i = 1;                           //log(base2) of N points= # of stages
 //Liczymy ilosc stopni mnoen motylkowych
 do
  {
   num_stages +=1;
   i = i*2;
  }while (i<N);
 leg_diff = N/2; 		     	    //difference between upper&lower legs
 step = 1;//freqStabilizer/N;                   //step between values in twiddleW
 for (i = 0;i < num_stages; i++)  //for N-point FFT
  {
   index = 0;
   for (j = 0; j < leg_diff; j++)
    {
     //Pojedynczy motylek wyglda tak:
     for (upper_leg = j; upper_leg < N; upper_leg += (2*leg_diff))
      {
       lower_leg = upper_leg+leg_diff;
	 temp1.real = (Y[upper_leg]).real + (Y[lower_leg]).real;
       temp1.imag = (Y[upper_leg]).imag + (Y[lower_leg]).imag;
       temp2.real = (Y[upper_leg]).real - (Y[lower_leg]).real;
       temp2.imag = (Y[upper_leg]).imag - (Y[lower_leg]).imag;
       (Y[lower_leg]).real = temp2.real*(twiddle_W[index]).real
				    -temp2.imag*(twiddle_W[index]).imag;
       (Y[lower_leg]).imag = temp2.real*(twiddle_W[index]).imag
				    +temp2.imag*(twiddle_W[index]).real;
       (Y[upper_leg]).real = temp1.real;
       (Y[upper_leg]).imag = temp1.imag;
      }
   index += step;
    }
     leg_diff = leg_diff/2;
     step *= 2;
  }
  j = 0;
  for (i = 1; i < (N-1); i++)     //bit reversal for resequencing data
    {
     k = N/2;
     while (k <= j)
      {
       j = j - k;
       k = k/2;
      }
     j = j + k;
     if (i<j)
      {
       temp1.real = (Y[j]).real;
       temp1.imag = (Y[j]).imag;
       (Y[j]).real = (Y[i]).real;
       (Y[j]).imag = (Y[i]).imag;
       (Y[i]).real = temp1.real;
       (Y[i]).imag = temp1.imag;
      }
    }
   return;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FFTBitmap(COMPLEX *YR,COMPLEX *YG,COMPLEX *YB,Graphics::TBitmap *src)
//Robi FFT, zapisuje j do YR,YG,YB
{
  //Wyznaczamy wysoko i szeroko dla FFT (musi by 2^n)
  uint fftWidth=1;
  uint fftHeight=1;
  uint fWidthMul=0;
  uint fHeightMul=0;
  uint srcWidth=src->Width;
  uint srcHeight=src->Height;
  do
    {
    fftWidth*=2;
    fWidthMul++;
    } while (fftWidth<srcWidth);
  do
    {
    fftHeight*=2;
    fHeightMul++;
    } while (fftHeight<srcHeight);
  //Do paska stanu
  dWidth=fftWidth<<1;
  dHeight=fftHeight;
  //Aktualizujemy pasek stanu
  AdvanceMax=Extended(dWidth)*dHeight;
  i=0;j=0;
  Synchronize(UpdateAdvancePos);
  //Dopasowujemy sobie parametry obrazkw
  src->PixelFormat=pf24bit;
  //Allokujemy sobie fufory
  twiddle_W=new COMPLEX[max(fftWidth,fftHeight)];
  //Ustawiamy zmienn twiddleW do obsugi szerokoci
  uint freqStabilizer=fftWidth;
  for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
   {
   twiddle_W[fftI].real = cos(M_PI*2.0*fftI/freqStabilizer); //Re component of twiddle constants
   twiddle_W[fftI].imag =-sin(M_PI*2.0*fftI/freqStabilizer); //Im component of twiddle constants
   }
 //I liczymy FFT wierszy
 uint fftJ;
 for (fftJ = 0 ; fftJ<srcHeight ; fftJ++)
   {
   //Aktualizujemy pasek stanu
   i=fftJ>>2;
   if (Terminated) return;
   Synchronize(UpdateAdvancePos);
   //Wypeniamy
   for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
     {
     //Pobieramy obraz rdowy do YR,YG,YB
     if (fftI<srcWidth)
       {
       YR[(fftJ<<fWidthMul)+fftI].real = ((uchar *)src->ScanLine[fftJ])[fftI*3  ];    //real components
       YG[(fftJ<<fWidthMul)+fftI].real = ((uchar *)src->ScanLine[fftJ])[fftI*3+1];    //real components
       YB[(fftJ<<fWidthMul)+fftI].real = ((uchar *)src->ScanLine[fftJ])[fftI*3+2];    //real components
       }
      else
       {
       uint idx=(2*srcWidth-fftI-1);
       YR[(fftJ<<fWidthMul)+fftI].real = ((uchar *)src->ScanLine[fftJ])[idx*3  ];    //real components
       YG[(fftJ<<fWidthMul)+fftI].real = ((uchar *)src->ScanLine[fftJ])[idx*3+1];    //real components
       YB[(fftJ<<fWidthMul)+fftI].real = ((uchar *)src->ScanLine[fftJ])[idx*3+2];    //real components
       }
     YR[(fftJ<<fWidthMul)+fftI].imag = 0.0;	    //imag components = 0
     YG[(fftJ<<fWidthMul)+fftI].imag = 0.0;	    //imag components = 0
     YB[(fftJ<<fWidthMul)+fftI].imag = 0.0;	    //imag components = 0
     }
   //Robimy FFT wierszy
   FFT(YR+(fftJ<<fWidthMul), fftWidth);
   FFT(YG+(fftJ<<fWidthMul), fftWidth);
   FFT(YB+(fftJ<<fWidthMul), fftWidth);
   }
 //Dorabiamy wiersze uzupeniajce od srcHeight do fftHeight
 if (fftJ<fftHeight)
    do
     {
     //Aktualizujemy pasek stanu
     i=fftJ>>2;
     if (Terminated) return;
     Synchronize(UpdateAdvancePos);
     //I liczymy
     uint idxJ=(2*srcHeight-fftJ-1);
     for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
       {
       YR[(fftJ<<fWidthMul)+fftI] = YR[(idxJ<<fWidthMul)+fftI];
       YG[(fftJ<<fWidthMul)+fftI] = YG[(idxJ<<fWidthMul)+fftI];
       YB[(fftJ<<fWidthMul)+fftI] = YB[(idxJ<<fWidthMul)+fftI];
       }
     fftJ++;
     }
    while (fftJ<fftHeight);
  //Wiersze gotowe - transformujemy kolumny
  //Ustawiamy zmienn twiddleW do obsugi wysokoci
  freqStabilizer=fftHeight;
  for (uint fftI = 0 ; fftI<fftHeight ; fftI++)
   {
   twiddle_W[fftI].real = cos(M_PI*2.0*fftI/freqStabilizer); //Re component of twiddle constants
   twiddle_W[fftI].imag =-sin(M_PI*2.0*fftI/freqStabilizer); //Im component of twiddle constants
   }
 //I lecimy po kolumnach
 COMPLEX *bufR=new COMPLEX[fftHeight];
 COMPLEX *bufG=new COMPLEX[fftHeight];
 COMPLEX *bufB=new COMPLEX[fftHeight];
 for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
   {
   //Aktualizujemy pasek stanu
   i=(fftHeight>>2);
   j=(fftI<<(fHeightMul-1));
   if (Terminated) return;
   Synchronize(UpdateAdvancePos);
   //Uzupeniamy bufory
   for (uint fftJ = 0 ; fftJ<fftHeight ; fftJ++)
     {
     bufR[fftJ] = YR[(fftJ<<fWidthMul)+fftI];
     bufG[fftJ] = YG[(fftJ<<fWidthMul)+fftI];
     bufB[fftJ] = YB[(fftJ<<fWidthMul)+fftI];
     }
   //Liczymy FFT
   FFT(bufR, fftHeight);
   FFT(bufG, fftHeight);
   FFT(bufB, fftHeight);
   //Bufory z powrotem do macierzy
   for (uint fftJ = 0 ; fftJ<fftHeight ; fftJ++)
     {
     YR[(fftJ<<fWidthMul)+fftI]=bufR[fftJ];
     YG[(fftJ<<fWidthMul)+fftI]=bufG[fftJ];
     YB[(fftJ<<fWidthMul)+fftI]=bufB[fftJ];
     }
   }
 delete bufR;
 delete bufG;
 delete bufB;

  //Czycimy bufory
  delete twiddle_W;
  twiddle_W=NULL;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FFTTable(COMPLEX *Y,Extended *src,uint srcWidth,uint srcHeight)
//Robi FFT, zapisuje j do Y
{
  //Wyznaczamy wysoko i szeroko dla FFT (musi by 2^n)
  uint fftWidth=1;
  uint fftHeight=1;
  uint fWidthMul=0;
  uint fHeightMul=0;
  do
    {
    fftWidth*=2;
    fWidthMul++;
    } while (fftWidth<srcWidth);
  do
    {
    fftHeight*=2;
    fHeightMul++;
    } while (fftHeight<srcHeight);
  //Do paska stanu
//  dWidth=fftWidth<<1;
//  dHeight=fftHeight;
  //Aktualizujemy pasek stanu
//  AdvanceMax=Extended(dWidth)*dHeight;
//  i=0;j=0;
//  Synchronize(UpdateAdvancePos);
  //Allokujemy sobie fufory
  twiddle_W=new COMPLEX[max(fftWidth,fftHeight)];
  //Ustawiamy zmienn twiddleW do obsugi szerokoci
  for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
   {
   twiddle_W[fftI].real = cos(M_PI*2.0*fftI/fftWidth); //Re component of twiddle constants
   twiddle_W[fftI].imag =-sin(M_PI*2.0*fftI/fftWidth); //Im component of twiddle constants
   }
 //I liczymy FFT wierszy
 uint fftJ;
 for (fftJ = 0 ; fftJ<srcHeight ; fftJ++)
   {
   //Aktualizujemy pasek stanu
//   i=fftJ>>2;
   if (Terminated) return;
//   Synchronize(UpdateAdvancePos);
   //Wypeniamy
   for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
     {
     //Pobieramy obraz rdowy do YR,YG,YB
     if (fftI<srcWidth)
       {
       Y[(fftJ<<fWidthMul)+fftI].real = src[(fftJ<<fWidthMul)+fftI];    //real components
       }
      else
       {
       uint idx=(2*srcWidth-fftI-1);
       Y[(fftJ<<fWidthMul)+fftI].real = src[(fftJ<<fWidthMul)+idx];
       }
     Y[(fftJ<<fWidthMul)+fftI].imag = 0.0;	    //imag components = 0
     }
   //Robimy FFT wierszy
   FFT(Y+(fftJ<<fWidthMul), fftWidth);
   }
 //Dorabiamy wiersze uzupeniajce od srcHeight do fftHeight
 if (fftJ<fftHeight)
    do
     {
     //Aktualizujemy pasek stanu
//     i=fftJ>>2;
     if (Terminated) return;
//     Synchronize(UpdateAdvancePos);
     //I liczymy
     uint idxJ=(2*srcHeight-fftJ-1);
     for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
       {
       Y[(fftJ<<fWidthMul)+fftI] = Y[(idxJ<<fWidthMul)+fftI];
       }
     fftJ++;
     }
    while (fftJ<fftHeight);
  //Wiersze gotowe - transformujemy kolumny
  //Ustawiamy zmienn twiddleW do obsugi wysokoci
  for (uint fftI = 0 ; fftI<fftHeight ; fftI++)
   {
   twiddle_W[fftI].real = cos(M_PI*2.0*fftI/fftHeight); //Re component of twiddle constants
   twiddle_W[fftI].imag =-sin(M_PI*2.0*fftI/fftHeight); //Im component of twiddle constants
   }
 //I lecimy po kolumnach
 COMPLEX *buf=new COMPLEX[fftHeight];
 for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
   {
   //Aktualizujemy pasek stanu
//   i=(fftHeight>>2);
//   j=(fftI<<(fHeightMul-1));
   if (Terminated) return;
//   Synchronize(UpdateAdvancePos);
   //Uzupeniamy bufory
   for (uint fftJ = 0 ; fftJ<fftHeight ; fftJ++)
     {
     buf[fftJ] = Y[(fftJ<<fWidthMul)+fftI];
     }
   //Liczymy FFT
   FFT(buf, fftHeight);
   //Bufory z powrotem do macierzy
   for (uint fftJ = 0 ; fftJ<fftHeight ; fftJ++)
     {
     Y[(fftJ<<fWidthMul)+fftI]=buf[fftJ];
     }
   }
 delete buf;

  //Czycumy bufory
  delete twiddle_W;
  twiddle_W=NULL;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::RFTBitmap(Graphics::TBitmap *dest,uint srcWidth,uint srcHeight,COMPLEX *YR,COMPLEX *YG,COMPLEX *YB)
//Robi Reverse FFT z YR,YG,YB do dest
{
  //Wyznaczamy wysoko i szeroko dla FFT (musi by 2^n)
  uint fftWidth=1;
  uint fftHeight=1;
  uint fWidthMul=0;
  uint fHeightMul=0;
  do
    {
    fftWidth*=2;
    fWidthMul++;
    } while (fftWidth<srcWidth);
  do
    {
    fftHeight*=2;
    fHeightMul++;
    } while (fftHeight<srcHeight);
  //Aktualizujemy pasek stanu
  dWidth=(fftWidth<<1);
  dHeight=fftHeight;
  AdvanceMax=Extended(dWidth)*dHeight;
  i=fftWidth>>1;j=0;
  Synchronize(UpdateAdvancePos);
  //Dopasowujemy sobie parametry obrazkw
  dest->PixelFormat=pf24bit;
  dest->Width=srcWidth;
  dest->Height=srcHeight;
  //Allokujemy sobie fufory
  twiddle_W=new COMPLEX[max(fftWidth,fftHeight)];
  //Transformujemy kolumny
  //Ustawiamy zmienn twiddleW do obsugi wysokoci
  uint freqStabilizer=fftHeight;
  for (uint fftI = 0 ; fftI<fftHeight ; fftI++)
   {
   twiddle_W[fftI].real = cos(M_PI*2.0*fftI/freqStabilizer); //Re component of twiddle constants
   twiddle_W[fftI].imag = sin(M_PI*2.0*fftI/freqStabilizer); //Im component of twiddle constants
   }
 //I lecimy po kolumnach
 COMPLEX *bufR=new COMPLEX[fftHeight];
 COMPLEX *bufG=new COMPLEX[fftHeight];
 COMPLEX *bufB=new COMPLEX[fftHeight];
 for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
   {
   //Aktualizujemy pasek stanu
   i=(fftHeight>>1);
   j=(fftI<<(fHeightMul-1));
   if (Terminated) return;
   Synchronize(UpdateAdvancePos);
   //Uzupeniamy bufory
   for (uint fftJ = 0 ; fftJ<fftHeight ; fftJ++)
     {
     bufR[fftJ] = YR[(fftJ<<fWidthMul)+fftI];
     bufG[fftJ] = YG[(fftJ<<fWidthMul)+fftI];
     bufB[fftJ] = YB[(fftJ<<fWidthMul)+fftI];
     }
   //Liczymy RFFT
   FFT(bufR, fftHeight);
   FFT(bufG, fftHeight);
   FFT(bufB, fftHeight);
   //Bufory z powrotem do macierzy
   for (uint fftJ = 0 ; fftJ<fftHeight ; fftJ++)
     {
     YR[(fftJ<<fWidthMul)+fftI]=bufR[fftJ];
     YG[(fftJ<<fWidthMul)+fftI]=bufG[fftJ];
     YB[(fftJ<<fWidthMul)+fftI]=bufB[fftJ];
     }
   }
 delete bufR;
 delete bufG;
 delete bufB;
  //Transformujemy wiersze
  //Ustawiamy zmienn twiddleW do obsugi szerokoci
  freqStabilizer=fftWidth;
  for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
   {
   twiddle_W[fftI].real = cos(M_PI*2.0*fftI/freqStabilizer); //Re component of twiddle constants
   twiddle_W[fftI].imag = sin(M_PI*2.0*fftI/freqStabilizer); //Im component of twiddle constants
   }
 //Poprawka paska stanu
 i=(fftHeight>>1)+(fftHeight>>2);
 //I liczymy FFT wierszy
 for (uint fftJ = 0 ; fftJ<srcHeight ; fftJ++)
   {
   //Aktualizujemy pasek stanu
   j=fftJ*(fftWidth>>1);
   if (Terminated) return;
   Synchronize(UpdateAdvancePos);
   //Robimy RFFT wierszy
   FFT(YR+(fftJ<<fWidthMul), fftWidth);
   FFT(YG+(fftJ<<fWidthMul), fftWidth);
   FFT(YB+(fftJ<<fWidthMul), fftWidth);
   //Wsadzamy wartoci do bitmapy docelowej
   for (uint fftI = 0 ; fftI<srcWidth ; fftI++)
     {
     ((uchar *)dest->ScanLine[fftJ])[fftI*3  ]=Ceil(min(max(((YR[(fftJ<<fWidthMul)+fftI].real+0.49)/(fftHeight<<fWidthMul)),0),255));
     ((uchar *)dest->ScanLine[fftJ])[fftI*3+1]=Ceil(min(max(((YG[(fftJ<<fWidthMul)+fftI].real+0.49)/(fftHeight<<fWidthMul)),0),255));
     ((uchar *)dest->ScanLine[fftJ])[fftI*3+2]=Ceil(min(max(((YB[(fftJ<<fWidthMul)+fftI].real+0.49)/(fftHeight<<fWidthMul)),0),255));
     }
   }
  //Czycumy bufory
  delete twiddle_W;
  twiddle_W=NULL;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductFFTBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src)
{
  //Wyznaczamy wysoko i szeroko dla FFT (musi by 2^n)
  uint fftWidth=1;
  uint fftHeight=1;
  uint fWidthMul=0;
  uint fHeightMul=0;
  uint srcWidth=src->Width;
  uint srcHeight=src->Height;
  do
    {
    fftWidth*=2;
    fWidthMul++;
    } while (fftWidth<srcWidth);
  do
    {
    fftHeight*=2;
    fHeightMul++;
    } while (fftHeight<srcHeight);
  //Dopasowujemy sobie parametry obrazkw
  src->PixelFormat=pf24bit;
  dest->PixelFormat=pf24bit;
  dest->Width=fftWidth;
  dest->Height=fftHeight;
  //Allokujemy sobie fufory
  COMPLEX *YR=new COMPLEX[fftWidth*fftHeight];
  COMPLEX *YG=new COMPLEX[fftWidth*fftHeight];
  COMPLEX *YB=new COMPLEX[fftWidth*fftHeight];
  //Liczymy FFT
  FFTBitmap(YR,YG,YB,src);

  SwitchQuartersInTable(YR,fftWidth,fftHeight);
  SwitchQuartersInTable(YG,fftWidth,fftHeight);
  SwitchQuartersInTable(YB,fftWidth,fftHeight);

//  RFTBitmap(destR,srcWidth,srcHeight,YR,YG,YB);

   //Zwracamy wartoci
  Extended DivFactor=(fftHeight<<(fWidthMul));
  for (uint fftJ = 0 ; fftJ<fftHeight ; fftJ++)
    {
    for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
      {
      Extended x=(YR[(fftJ<<fWidthMul)+fftI].real/DivFactor + 128);
      ((uchar *)dest->ScanLine[fftJ])[fftI*3  ]=Ceil(min(max(x,0),255));
      x=(YG[(fftJ<<fWidthMul)+fftI].real/DivFactor + 128);
      ((uchar *)dest->ScanLine[fftJ])[fftI*3+1]=Ceil(min(max(x,0),255));
      x=(YB[(fftJ<<fWidthMul)+fftI].real/DivFactor + 128);
      ((uchar *)dest->ScanLine[fftJ])[fftI*3+2]=Ceil(min(max(x,0),255));
//     ((uchar *)destI->ScanLine[fftJ])[fftI*3  ]==YR[fftI].imag;
//     ((uchar *)destI->ScanLine[fftJ])[fftI*3+1]==YG[fftI].imag;
//     ((uchar *)destI->ScanLine[fftJ])[fftI*3+2]==YB[fftI].imag;
      }
    }

/*
  //Wypeniamy bufor - bedzie potrzebny do odtworzenia czci urojonej
  COMPLEX *re_to_im_W=new COMPLEX[fftWidth];
  COMPLEX *re_to_im_H=new COMPLEX[fftHeight];
  for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
   {
   re_to_im_W[fftI].real = cos(M_PI*2.0*fftI/fftWidth); //Re component of twiddle constants
   re_to_im_W[fftI].imag = sin(M_PI*2.0*fftI/fftWidth); //Im component of twiddle constants
   }
  for (uint fftI = 0 ; fftI<fftHeight ; fftI++)
   {
   re_to_im_H[fftI].real = cos(M_PI*2.0*fftI/fftHeight); //Re component of twiddle constants
   re_to_im_H[fftI].imag = sin(M_PI*2.0*fftI/fftHeight); //Im component of twiddle constants
   }
  for (uint fftJ = 0 ; fftJ<srcHeight ; fftJ++)
    {
    for (uint fftI = 0 ; fftI<srcWidth ; fftI++)
      {
      Extended valR=((uchar *)dest->ScanLine[fftJ])[fftI*3  ];
      Extended valG=((uchar *)dest->ScanLine[fftJ])[fftI*3+1];
      Extended valB=((uchar *)dest->ScanLine[fftJ])[fftI*3+2];
      valR=(valR-128)*DivFactor;
      valG=(valG-128)*DivFactor;
      valB=(valB-128)*DivFactor;

      YR[(fftJ<<fWidthMul)+fftI].real=valR;
      YG[(fftJ<<fWidthMul)+fftI].real=valG;
      YB[(fftJ<<fWidthMul)+fftI].real=valB;
      Extended ReToImMultip=(-(re_to_im_H[fftJ].imag*re_to_im_W[fftI].real) - (re_to_im_W[fftI].imag*re_to_im_H[fftJ].real));

      {
      Extended TmpDivider=((re_to_im_H[fftJ].real*re_to_im_W[fftI].real) - (re_to_im_W[fftI].imag*re_to_im_H[fftJ].imag));
//      if (TmpDivider>0)
//          ReToImMultip/=max(TmpDivider,0.1);
//        else
//          ReToImMultip/=min(TmpDivider,-0.1);
      }

      valR*=ReToImMultip;
      valG*=ReToImMultip;
      valB*=ReToImMultip;
      YR[(fftJ<<fWidthMul)+fftI].imag=valR;
      YG[(fftJ<<fWidthMul)+fftI].imag=valG;
      YB[(fftJ<<fWidthMul)+fftI].imag=valB;
      }
    }
  //Czyscimy bufory
  delete re_to_im_W;
  delete re_to_im_H;
  //Liczymy RFFT
  RFTBitmap(dest,srcWidth,srcHeight,YR,YG,YB);
*/
  //Czycumy bufory
  delete YR;
  delete YG;
  delete YB;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductRFTBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src)
{
  //Wyznaczamy wysoko i szeroko dla FFT (musi by 2^n)
  uint fftWidth=1;
  uint fftHeight=1;
  uint fWidthMul=0;
  uint fHeightMul=0;
  uint srcWidth=src->Width;
  uint srcHeight=src->Height;
  do
    {
    fftWidth*=2;
    fWidthMul++;
    } while (fftWidth<srcWidth);
  do
    {
    fftHeight*=2;
    fHeightMul++;
    } while (fftHeight<srcHeight);
  //Dopasowujemy sobie parametry obrazkw
  src->PixelFormat=pf24bit;
  dest->PixelFormat=pf24bit;
  dest->Width=fftWidth;
  dest->Height=fftHeight;

  Extended DivFactor=(fftHeight<<(fWidthMul));
  //Allokujemy sobie fufory
  COMPLEX *YR=new COMPLEX[fftWidth*fftHeight];
  COMPLEX *YG=new COMPLEX[fftWidth*fftHeight];
  COMPLEX *YB=new COMPLEX[fftWidth*fftHeight];
  //Wypeniamy bufor - bedzie potrzebny do odtworzenia czci urojonej
  COMPLEX *re_to_im_W=new COMPLEX[fftWidth];
  COMPLEX *re_to_im_H=new COMPLEX[fftHeight];
  for (uint fftI = 0 ; fftI<fftWidth ; fftI++)
   {
   re_to_im_W[fftI].real = cos(M_PI*2.0*fftI/fftWidth); //Re component of twiddle constants
   re_to_im_W[fftI].imag = sin(M_PI*2.0*fftI/fftWidth); //Im component of twiddle constants
   }
  for (uint fftI = 0 ; fftI<fftHeight ; fftI++)
   {
   re_to_im_H[fftI].real = cos(M_PI*2.0*fftI/fftHeight); //Re component of twiddle constants
   re_to_im_H[fftI].imag = sin(M_PI*2.0*fftI/fftHeight); //Im component of twiddle constants
   }
  for (uint fftJ = 0 ; fftJ<srcHeight ; fftJ++)
    {
    for (uint fftI = 0 ; fftI<srcWidth ; fftI++)
      {
      Extended valR=((uchar *)src->ScanLine[fftJ])[fftI*3  ];
      Extended valG=((uchar *)src->ScanLine[fftJ])[fftI*3+1];
      Extended valB=((uchar *)src->ScanLine[fftJ])[fftI*3+2];
      valR=(valR-128)*DivFactor;
      valG=(valG-128)*DivFactor;
      valB=(valB-128)*DivFactor;

      YR[(fftJ<<fWidthMul)+fftI].real=valR;
      YG[(fftJ<<fWidthMul)+fftI].real=valG;
      YB[(fftJ<<fWidthMul)+fftI].real=valB;
      Extended ReToImMultip=(-(re_to_im_H[fftJ].imag*re_to_im_W[fftI].real) - (re_to_im_W[fftI].imag*re_to_im_H[fftJ].real));


      {
//      Extended TmpDivider=((re_to_im_H[fftJ].real*re_to_im_W[fftI].real) - (re_to_im_W[fftI].imag*re_to_im_H[fftJ].imag));
//      if (TmpDivider>0)
//          ReToImMultip/=max(TmpDivider,0.1);
//        else
//          ReToImMultip/=min(TmpDivider,-0.1);
      }

      valR*=ReToImMultip;
      valG*=ReToImMultip;
      valB*=ReToImMultip;
      YR[(fftJ<<fWidthMul)+fftI].imag=valR;
      YG[(fftJ<<fWidthMul)+fftI].imag=valG;
      YB[(fftJ<<fWidthMul)+fftI].imag=valB;
      }
    }
  //Czyscimy bufory
  delete re_to_im_W;
  delete re_to_im_H;
  //Zamiana wiartek
  SwitchQuartersInTable(YR,fftWidth,fftHeight);
  SwitchQuartersInTable(YG,fftWidth,fftHeight);
  SwitchQuartersInTable(YB,fftWidth,fftHeight);
  //Liczymy RFFT
  RFTBitmap(dest,srcWidth,srcHeight,YR,YG,YB);

  //Czycimy bufory
  delete YR;
  delete YG;
  delete YB;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductMultiplyTable(COMPLEX *dest,uint destWidth,uint destHeight,COMPLEX *filter,uint filterWidth,uint filterHeight)
{
  uint filterHalfWidth=filterWidth>>2;
  uint filterHalfHeight=filterHeight>>2;
  Extended Scale=1;

  for (uint fftI = 0 ; fftI<filterHalfHeight ; fftI++)
    {
    for (uint fftJ = 0 ; fftJ<filterHalfWidth ; fftJ++)
      {
      //Grny lewy
      {
      Extended RealMul=dest[fftI*destWidth+fftJ].real*filter[(fftI)*filterWidth+(fftJ)].real;
      Extended ImagMul=dest[fftI*destWidth+fftJ].imag*filter[(fftI)*filterWidth+(fftJ)].imag;
      Extended RIMul=dest[fftI*destWidth+fftJ].real*filter[(fftI)*filterWidth+(fftJ)].imag;
      Extended IRMul=dest[fftI*destWidth+fftJ].imag*filter[(fftI)*filterWidth+(fftJ)].real;
      dest[fftI*destWidth+fftJ].real=(RealMul-ImagMul)*Scale;
      dest[fftI*destWidth+fftJ].imag=(RIMul+IRMul)*Scale;
      }
      //Grny prawy
      {
      Extended RealMul=dest[fftI*destWidth+(destWidth-fftJ-1)].real*filter[(fftI)*filterWidth+(filterWidth-fftJ-1)].real;
      Extended ImagMul=dest[fftI*destWidth+(destWidth-fftJ-1)].imag*filter[(fftI)*filterWidth+(filterWidth-fftJ-1)].imag;
      Extended RIMul=dest[fftI*destWidth+(destWidth-fftJ-1)].real*filter[(fftI)*filterWidth+(filterWidth-fftJ-1)].imag;
      Extended IRMul=dest[fftI*destWidth+(destWidth-fftJ-1)].imag*filter[(fftI)*filterWidth+(filterWidth-fftJ-1)].real;
      dest[fftI*destWidth+(destWidth-fftJ-1)].real=(RealMul-ImagMul)*Scale;
      dest[fftI*destWidth+(destWidth-fftJ-1)].imag=(RIMul+IRMul)*Scale;
      }
      //Dolny lewy
      {
      Extended RealMul=dest[(destHeight-fftI-1)*destWidth+fftJ].real*filter[(filterHeight-fftI-1)*filterWidth+(fftJ)].real;
      Extended ImagMul=dest[(destHeight-fftI-1)*destWidth+fftJ].imag*filter[(filterHeight-fftI-1)*filterWidth+(fftJ)].imag;
      Extended RIMul=dest[(destHeight-fftI-1)*destWidth+fftJ].real*filter[(filterHeight-fftI-1)*filterWidth+(fftJ)].imag;
      Extended IRMul=dest[(destHeight-fftI-1)*destWidth+fftJ].imag*filter[(filterHeight-fftI-1)*filterWidth+(fftJ)].real;
      dest[(destHeight-fftI-1)*destWidth+fftJ].real=(RealMul-ImagMul)*Scale;
      dest[(destHeight-fftI-1)*destWidth+fftJ].imag=(RIMul+IRMul)*Scale;
      }
      //Dolny prawy
      {
      Extended RealMul=dest[(destHeight-fftI-1)*destWidth+(destWidth-fftJ-1)].real*filter[(filterHeight-fftI-1)*filterWidth+(filterWidth-fftJ-1)].real;
      Extended ImagMul=dest[(destHeight-fftI-1)*destWidth+(destWidth-fftJ-1)].imag*filter[(filterHeight-fftI-1)*filterWidth+(filterWidth-fftJ-1)].imag;
      Extended RIMul=dest[(destHeight-fftI-1)*destWidth+(destWidth-fftJ-1)].real*filter[(filterHeight-fftI-1)*filterWidth+(filterWidth-fftJ-1)].imag;
      Extended IRMul=dest[(destHeight-fftI-1)*destWidth+(destWidth-fftJ-1)].imag*filter[(filterHeight-fftI-1)*filterWidth+(filterWidth-fftJ-1)].real;
      dest[(destHeight-fftI-1)*destWidth+(destWidth-fftJ-1)].real=(RealMul-ImagMul)*Scale;
      dest[(destHeight-fftI-1)*destWidth+(destWidth-fftJ-1)].imag=(RIMul+IRMul)*Scale;
      }

      }
    }
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductMultipSameTbl(COMPLEX *dest,COMPLEX *filter,uint filterWidth,uint filterHeight)
{
  for (uint fftI = 0 ; fftI<filterHeight ; fftI++)
    {
    for (uint fftJ = 0 ; fftJ<filterWidth ; fftJ++)
      {
      uint srcIdx=fftI*filterWidth+fftJ;
      Extended RealMul=dest[srcIdx].real*filter[srcIdx].real;
      Extended ImagMul=dest[srcIdx].imag*filter[srcIdx].imag;
      Extended RIMul=dest[srcIdx].real*filter[srcIdx].imag;
      Extended IRMul=dest[srcIdx].imag*filter[srcIdx].real;
      dest[srcIdx].real=(RealMul-ImagMul);
      dest[srcIdx].imag=(RIMul+IRMul);
      }
    }

}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::SwitchQuartersInTable(COMPLEX *dest,uint destWidth,uint destHeight)
{
  uint destWidthDiv2=destWidth/2;
  uint destHeightDiv2=destHeight/2;
  for (uint fftI = 0 ; fftI<destHeightDiv2 ; fftI++)
    {
    for (uint fftJ = 0 ; fftJ<destWidthDiv2 ; fftJ++)
      {
      {//Pierwsza cwiardka z czwart
      uint srcIdx=fftI*destWidth+fftJ;
      uint dstIdx=(destHeightDiv2)*destWidth+(destWidthDiv2)+srcIdx;
      COMPLEX aValue=dest[srcIdx];
      dest[srcIdx]=dest[dstIdx];
      dest[dstIdx]=aValue;
      }
      {//Druga cwiardka z trzeci
      uint srcIdx=(fftI)*destWidth+(destWidthDiv2+fftJ);
      uint dstIdx=(destHeightDiv2+fftI)*destWidth+(fftJ);
      COMPLEX aValue=dest[srcIdx];
      dest[srcIdx]=dest[dstIdx];
      dest[dstIdx]=aValue;
      }
/*
      {//Pierwsza cwiardka
      uint srcIdx=fftI*destWidth+fftJ;
      uint dstIdx=(destHeightDiv2-fftI-1)*destWidth+(destWidthDiv2-fftJ-1);
      COMPLEX aValue=dest[srcIdx];
      dest[srcIdx]=dest[dstIdx];
      dest[dstIdx]=aValue;
      }
      {//Druga cwiardka
      uint srcIdx=fftI*destWidth+(destHeight-fftJ-1);
      uint dstIdx=(destHeightDiv2-fftI-1)*destWidth+(destWidthDiv2+fftJ);
      COMPLEX aValue=dest[srcIdx];
      dest[srcIdx]=dest[dstIdx];
      dest[dstIdx]=aValue;
      }
      {//Trzecia cwiardka
      uint srcIdx=(destHeight-fftI-1)*destWidth+(fftJ);
      uint dstIdx=(destHeightDiv2+fftI)*destWidth+(destWidthDiv2-fftJ-1);
      COMPLEX aValue=dest[srcIdx];
      dest[srcIdx]=dest[dstIdx];
      dest[dstIdx]=aValue;
      }
      {//Czwarta cwiardka
      uint srcIdx=(destHeight-fftI-1)*destWidth+(destWidth-fftJ-1);
      uint dstIdx=(destHeightDiv2+fftI)*destWidth+(destWidthDiv2+fftJ);
      COMPLEX aValue=dest[srcIdx];
      dest[srcIdx]=dest[dstIdx];
      dest[dstIdx]=aValue;
      }*/
      }
    }
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::SwitchQuartersInBitmap(Graphics::TBitmap *dest,uint destWidth,uint destHeight)
{
  uint destWidthDiv2=destWidth/2;
  uint destHeightDiv2=destHeight/2;
  uchar *buf=new uchar[(destWidthDiv2+1)*3];
  for (uint fftI = 0 ; fftI<destHeightDiv2 ; fftI++)
    {
    void *ptrA=dest->ScanLine[fftI];
    void *ptrB=(uchar *)dest->ScanLine[destHeightDiv2+fftI]+(destWidthDiv2)*3;
    memcpy(buf,ptrA,destWidthDiv2*3);
    memcpy(ptrA,ptrB,destWidthDiv2*3);
    memcpy(ptrB,buf,destWidthDiv2*3);

    ptrA=(uchar *)dest->ScanLine[fftI]+(destWidthDiv2)*3;
    ptrB=(uchar *)dest->ScanLine[destHeightDiv2+fftI];
    memcpy(buf,ptrA,destWidthDiv2*3);
    memcpy(ptrA,ptrB,destWidthDiv2*3);
    memcpy(ptrB,buf,destWidthDiv2*3);
    }
  delete buf;
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::FFTGaussianBlurBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended level)
{
  if ((dest==NULL)||(src==NULL)) return;
  //Wyznaczamy wysoko i szeroko dla FFT (musi by 2^n)
  uint fftWidth=1;
  uint fftHeight=1;
  uint fWidthMul=0;
  uint fHeightMul=0;
  uint srcWidth=src->Width;
  uint srcHeight=src->Height;
  do
    {
    fftWidth*=2;
    fWidthMul++;
    } while (fftWidth<srcWidth);
  do
    {
    fftHeight*=2;
    fHeightMul++;
    } while (fftHeight<srcHeight);
  //Dopasowujemy sobie parametry obrazkw
  src->PixelFormat=pf24bit;
  dest->PixelFormat=pf24bit;
  dest->Width=fftWidth;
  dest->Height=fftHeight;
  //Allokujemy sobie fufory
  COMPLEX *YR=new COMPLEX[fftWidth*fftHeight];
  COMPLEX *YG=new COMPLEX[fftWidth*fftHeight];
  COMPLEX *YB=new COMPLEX[fftWidth*fftHeight];
  //Liczymy FFT
  FFTBitmap(YR,YG,YB,src);
  //Ustawiamy wiardki na odpowiednich miejscach
  //SwitchQuartersInTable(YR,fftWidth,fftHeight);
  //SwitchQuartersInTable(YG,fftWidth,fftHeight);
  //SwitchQuartersInTable(YB,fftWidth,fftHeight);

  //Tworzymy tablic filtru
  COMPLEX *filterF;
  {
  Extended StromoscMod=logl(stromoscG)/((level+1)*(level+1));
  Extended *filter=new Extended[fftWidth*fftHeight];
//  FillTableWithGauss2D(filter,fftWidth,fftHeight,StromoscMod,StromoscMod,(0.5*(Extended(srcWidth)-1.0)),(0.5*(Extended(srcHeight)-1.0)));
  FillTableWithGauss2D(filter,fftWidth,fftHeight,StromoscMod,StromoscMod,(0.5*(Extended(fftWidth)-1.0)),(0.5*(Extended(fftWidth)-1.0)));
  //Mamy filtr, ale do splotu potrzebujemy jego transformat
  filterF=new COMPLEX[fftWidth*fftHeight];
//  FFTTable(filterF,filter,fftWidth,fftHeight);

//ZAMIAST LICZY FFT FILTRU, PRZEPISUJEMY WSPӣCZYNNIKI REAL ZA, IMAG=1
  for (uint nI = 0 ; nI<fftHeight ; nI++)
    for (uint nJ = 0 ; nJ<fftWidth ; nJ++)
      {
      uint srcIdx=nI*fftWidth+nJ;
      filterF[srcIdx].real=(100*filter[srcIdx]);
      filterF[srcIdx].imag=(100*filter[srcIdx]);
      }
  SwitchQuartersInTable(filterF,fftWidth,fftHeight);

  //Ustawiamy wiardki na odpowiednich miejscach
  //SwitchQuartersInTable(filterF,fftWidth,fftHeight);
  delete filter;
  }
  //Tablica gotowa, pora na filtracj
  ConductMultipSameTbl(YR,filterF,fftWidth,fftHeight);
  ConductMultipSameTbl(YG,filterF,fftWidth,fftHeight);
  ConductMultipSameTbl(YB,filterF,fftWidth,fftHeight);
//  ConductMultiplyTable(YR,fftWidth,fftHeight,filterF,fftWidth,fftHeight);
//  ConductMultiplyTable(YG,fftWidth,fftHeight,filterF,fftWidth,fftHeight);
//  ConductMultiplyTable(YB,fftWidth,fftHeight,filterF,fftWidth,fftHeight);
  //I usuwamy filtr
  delete filterF;

  //Ustawiamy wiardki na odpowiednich miejscach
  //SwitchQuartersInTable(YR,fftWidth,fftHeight);
  //SwitchQuartersInTable(YG,fftWidth,fftHeight);
  //SwitchQuartersInTable(YB,fftWidth,fftHeight);
  //Liczymy odwrotne FFT
  RFTBitmap(dest,srcWidth,srcHeight,YR,YG,YB);
  //SwitchQuartersInBitmap(dest,fftWidth,fftHeight);
  //Czycimy bufory
  delete YR;
  delete YG;
  delete YB;

}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::CountHistogram(THistogramType HistType,Extended *destR,Extended *destG,Extended *destB,Graphics::TBitmap *src)
{
  //Czyszczenie tablic na histogram
  for (uint nI = 0 ; nI<256 ; nI++)
    {
    destR[nI]=0.0;
    destG[nI]=0.0;
    destB[nI]=0.0;
    }
  //Konwersja kolorw
  if (src->PixelFormat!=pf24bit)
      src->PixelFormat=pf24bit;
  //i liczymy histogram sumacyjny
  uint srcWidth=src->Width;
  uint srcHeight=src->Height;
  for (uint nI = 0 ; nI<srcHeight ; nI++)
    for (uint nJ = 0 ; nJ<srcWidth ; nJ++)
      {
      destB[((uchar *)src->ScanLine[nI])[nJ*3]]++;
      destG[((uchar *)src->ScanLine[nI])[nJ*3+1]]++;
      destR[((uchar *)src->ScanLine[nI])[nJ*3+2]]++;
/*    NIE UYWA! Bd Buildera5 - deallokuje si Handle na bitmap...
      Nie mona uywa Canvasa poza wtkiem gwnym
      uint pixel=src->Canvas->Pixels[nI][nJ];
      destR[(pixel)&255]++;
      destG[(pixel>>8)&255]++;
      destB[(pixel>>16)&255]++;
*/
      }
  if (HistType==htSimpleSum) return;
  if (HistType==htNormalized)
    {
    Extended sumR=0.0;
    Extended sumG=0.0;
    Extended sumB=0.0;
    for (uint nI = 0 ; nI<256 ; nI++)
      {
      sumR+=destR[nI];
      sumG+=destG[nI];
      sumB+=destB[nI];
      }
    for (uint nI = 0 ; nI<256 ; nI++)
      {
      destR[nI]/=sumR;
      destG[nI]/=sumG;
      destB[nI]/=sumB;
      }
    return;
    }
  if (HistType==htCummulative)
    {
    for (uint nI=1; nI < 256; nI++)
        {
        destR[nI] = destR[nI-1] + destR[nI];
        destG[nI] = destG[nI-1] + destG[nI];
        destB[nI] = destB[nI-1] + destB[nI];
        }
    return;
    }
}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductLevelsChange(Graphics::TBitmap *dest,uchar *curve,destChannel channel)
{
  int srcWidth=dest->Width;
  int srcHeight=dest->Height;
    dWidth=destPict->Width;
    dHeight=destPict->Height;
    AdvanceMax=Extended(dWidth)*dHeight;
  i=0;j=0;
  Synchronize(UpdateAdvancePos);
  if (dest->PixelFormat!=pf24bit)
      dest->PixelFormat=pf24bit;
  switch (channel)
    {
    case dcAll:
      {
      for (i = 0 ; i<srcHeight ; i++)
        {
        Synchronize(UpdateAdvancePos);
        for (j = 0 ; j<srcWidth ; j++)
          {
          uint x=((uchar *)dest->ScanLine[i])[j*3];
          ((uchar *)dest->ScanLine[i])[j*3]=curve[x];
          x=((uchar *)dest->ScanLine[i])[j*3+1];
          ((uchar *)dest->ScanLine[i])[j*3+1]=curve[x];
          x=((uchar *)dest->ScanLine[i])[j*3+2];
          ((uchar *)dest->ScanLine[i])[j*3+2]=curve[x];
          }
       }
      };break;
    case dcBlue:
      {
      for (i = 0 ; i<srcHeight ; i++)
        {
        Synchronize(UpdateAdvancePos);
        for (j = 0 ; j<srcWidth ; j++)
          {
          uint x=((uchar *)dest->ScanLine[i])[j*3];
          ((uchar *)dest->ScanLine[i])[j*3]=curve[x];
          }
        }
      };break;
    case dcGreen:
      {
      for (i = 0 ; i<srcHeight ; i++)
        {
        Synchronize(UpdateAdvancePos);
        for (j = 0 ; j<srcWidth ; j++)
          {
          uint x=((uchar *)dest->ScanLine[i])[j*3+1];
          ((uchar *)dest->ScanLine[i])[j*3+1]=curve[x];
          }
        }
      };break;
    case dcRed:
      {
      for (i = 0 ; i<srcHeight ; i++)
        {
        Synchronize(UpdateAdvancePos);
        for (j = 0 ; j<srcWidth ; j++)
          {
          uint x=((uchar *)dest->ScanLine[i])[j*3+2];
          ((uchar *)dest->ScanLine[i])[j*3+2]=curve[x];
          }
        }
      };break;
    }

}
//---------------------------------------------------------------------------
void TBitmapProcesorThread::ConductAutoLevels(Graphics::TBitmap *dest,destChannel channel)
{
  //Aktualizujemy zmienne
  int srcWidth=dest->Width;
  int srcHeight=dest->Height;
  dWidth=srcWidth;
  dHeight=srcHeight;
  AdvanceMax=Extended(dWidth)*dHeight;
  i=0;j=0;
  Synchronize(UpdateAdvancePos);
  if (dest->PixelFormat!=pf24bit)
      dest->PixelFormat=pf24bit;
    //Liczyny histogram skumulowany
    Extended *histR=new Extended[256];
    Extended *histG=new Extended[256];
    Extended *histB=new Extended[256];
    CountHistogram(htCummulative,histR,histG,histB,dest);
    if (histR[255] < 0.001) histR[255] = 0.001; // aby uniknac mozliwosci dzielenia przez 0
    if (histG[255] < 0.001) histG[255] = 0.001;
    if (histB[255] < 0.001) histB[255] = 0.001;
    //i zabieramy si za dopasowanie poziomw
  switch (channel)
    {
    case dcAll:
      {
      for (i = 0 ; i<srcHeight ; i++)
        {
        Synchronize(UpdateAdvancePos);
        for (j = 0 ; j<srcWidth ; j++)
          {
          uint srcPix;
          srcPix=((uchar *)dest->ScanLine[i])[j*3];
          ((uchar *)dest->ScanLine[i])[j*3] = 255.0*(histB[srcPix]/histB[255]);
          srcPix=((uchar *)dest->ScanLine[i])[j*3+1];
          ((uchar *)dest->ScanLine[i])[j*3+1] = 255.0*(histG[srcPix]/histG[255]);
          srcPix=((uchar *)dest->ScanLine[i])[j*3+2];
          ((uchar *)dest->ScanLine[i])[j*3+2] = 255.0*(histR[srcPix]/histR[255]);
          }
       }
      };break;
    case dcBlue:
      {
      for (i = 0 ; i<srcHeight ; i++)
        {
        Synchronize(UpdateAdvancePos);
        for (j = 0 ; j<srcWidth ; j++)
          {
          uint srcPix;
          srcPix=((uchar *)dest->ScanLine[i])[j*3];
          ((uchar *)dest->ScanLine[i])[j*3] = 255.0*(histB[srcPix]/histB[255]);
          }
       }
      };break;
    case dcGreen:
      {
      for (i = 0 ; i<srcHeight ; i++)
        {
        Synchronize(UpdateAdvancePos);
        for (j = 0 ; j<srcWidth ; j++)
          {
          uint srcPix;
          srcPix=((uchar *)dest->ScanLine[i])[j*3+1];
          ((uchar *)dest->ScanLine[i])[j*3+1] = 255.0*(histG[srcPix]/histG[255]);
          }
       }
      };break;
    case dcRed:
      {
      for (i = 0 ; i<srcHeight ; i++)
        {
        Synchronize(UpdateAdvancePos);
        for (j = 0 ; j<srcWidth ; j++)
          {
          uint srcPix;
          srcPix=((uchar *)dest->ScanLine[i])[j*3+2];
          Extended wynik=255.0*(histR[srcPix]/histR[255]);
          ((uchar *)dest->ScanLine[i])[j*3+2] =wynik;
          }
       }
      };break;
    }
}
//---------------------------------------------------------------------------

