//---------------------------------------------------------------------------

#ifndef BitmapProcesorUnitH
#define BitmapProcesorUnitH
//---------------------------------------------------------------------------
#include <Classes.hpp>
#include <jpeg.hpp>
#include <Graphics.hpp>
#include <Math.hpp>
#include <math.h>
#include <syncobjs.hpp>
#include <System.hpp>
#include "ExtendedGauge.hpp"
#include "FXProcs.h"
//---------------------------------------------------------------------------
	enum eFilter {
		kPoint	        = 0,
		kLinearInterp	= 1,
		kCubicInterp	= 2,
	};
//---------------------------------------------------------------------------
	enum destChannel {
		dcAll       = 0,
		dcRed	    = 1,
		dcGreen 	= 2,
		dcBlue  	= 3,
		dcNone  	= -1
	};
//---------------------------------------------------------------------------
	enum threadFunction
              {
		tfNothing           = 0x00,
		tfResizeBitmap      = 0x01,
		tfFlipBitmap        = 0x02,
                tfRotateBitmap      = 0x03,
                tfCreateHistBitmap  = 0x04,
                tfCurvLevelsBitmap  = 0x05,
                tfAutoLevelsBitmap  = 0x06,

		tfGaussianFilter  = 0x20,
		tfTriangularFilter= 0x21,
		tfSharpenFilter   = 0x22,
		tfEdgesFilter     = 0x23,
		tfExternalFilter  = 0x24,

		tfGaussianPaint   = 0x40,
		tfTriangularPaint = 0x41,
		tfAbstractPaint   = 0x42,
		tfSharpenPaint    = 0x43,
		tfEdgesPaint      = 0x44,

                tfFFTBitmap       = 0x60,
                tfRFTBitmap       = 0x61,
                tfFFTGaussFilter  = 0x62,
              };
//---------------------------------------------------------------------------
	enum THistogramType
              {
		htSimpleSum,  //Tylko sumuje
                htNormalized, //zwraca histogram znormalizowany 0..1
                htCummulative,//Histogram skumulowany

              };
//---------------------------------------------------------------------------
typedef struct {Extended real,imag;} COMPLEX;
//---------------------------------------------------------------------------
const Extended stromoscT=-0.1;
const Extended stromoscG=500;
//---------------------------------------------------------------------------
typedef struct {
  TLogPalette lPal;
  TPaletteEntry dummy[256];
} TLogPalEx;
//---------------------------------------------------------------------------
//FUNKCJE Z ASSEMBLERA (a_convolute.asm)
//#define USE_ASM
//#define USE_ASM_DYNAMIC
/*
struct TConvoluteFilterData
      {
	long m[9];
	long bias;
	void *dyna_func;
	DWORD dyna_size;
	DWORD dyna_old_protect;
	BOOL fClip;
      };

#define C_TOPOK		(1)
#define C_BOTTOMOK	(2)
#define C_LEFTOK	(4)
#define C_RIGHTOK	(8)

extern "C" void asm_convolute_run(
		void *dst,
		void *src,
		unsigned long width,
		unsigned long height,
		unsigned long srcstride,
		unsigned long dststride,
		const long *matrix);

extern "C" void asm_dynamic_convolute_run(
		void *dst,
		void *src,
		unsigned long width,
		unsigned long height,
		unsigned long srcstride,
		unsigned long dststride,
		const long *matrix,
		const void *dyna_func);
*/
//---------------------------------------------------------------------------
class TBitmapProcesorThread : public TThread
{
private:
        int i;
        int j;
        int dWidth;
        int dHeight;
        COMPLEX *twiddle_W;
protected:
        void __fastcall Execute();
        int filter_convolute_run(Graphics::TBitmap *destBmp,Graphics::TBitmap *srcBmp,Graphics::TBitmap *filter,Extended fWidthCenter,Extended fHeightCenter);
public:
        __fastcall TBitmapProcesorThread(bool CreateSuspended);
        __fastcall ~TBitmapProcesorThread();
        //Funkcje wtku
        void MakeMonochromeBitmapPalette(Graphics::TBitmap *dest);
        void MakeSignedRGBitmapPalette(Graphics::TBitmap *dest);
        void ResizeBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,eFilter fmode);
        void ResizeWithConvolution(Graphics::TBitmap *dest,Graphics::TBitmap *src,Graphics::TBitmap *filter);
        void ConductConvolutionUnsgn(Graphics::TBitmap *dest,Graphics::TBitmap *src,Graphics::TBitmap *filter);
        void ConductConvolutionUnsgn(Graphics::TBitmap *dest,Graphics::TBitmap *src,Graphics::TBitmap *filter,Extended fWidthCenter,Extended fHeightCenter);
        void ConductConvolutionSign(Graphics::TBitmap *dest,Graphics::TBitmap *src,Graphics::TBitmap *filter);
        void ConductConvolutionSign(Graphics::TBitmap *dest,Graphics::TBitmap *src,Graphics::TBitmap *filter,Extended fWidthCenter,Extended fHeightCenter);
        void ConductConvolutionSign(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended *filter,int fSize,Extended fDivider,Extended fAdder,bool UseAbsVal);
        void ConductConvolutionSign(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended *filter,int fSize,Extended fWidthCenter,Extended fHeightCenter,Extended fDivider,Extended fAdder,bool UseAbsVal);
        void GaussianBlurBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended level);
        void TriangularBlurBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended level);
        void SharpenBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended level,Extended size);
        void FindEdgesBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended level,Extended size);
        void FloatFilterBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended *filter,int fSize,Extended fDivider,Extended fAdder,bool UseAbsVal);
        //Filtry
        void FillBitmapWithGauss2D(Graphics::TBitmap *filter,Extended stromosc,Extended ffWidth,Extended ffHeight);
        void FillTableWithGauss2D(Extended *filter,uint fWidth,uint fHeight,Extended stromoscH,Extended stromoscW,Extended ffCenterW,Extended ffCenterH);
        void FillBitmapWithTriangle2D(Graphics::TBitmap *filter,Extended stromosc,Extended ffWidth,Extended ffHeight);
        void FillBitmapWithRotatedTriangle2D(Graphics::TBitmap *filter,Extended stromosc,Extended angle,Extended ffWidth,Extended ffHeight);
        void FillBitmapWithAbstract2D(Graphics::TBitmap *filter,int Number,Extended level,Extended ffWidth,Extended ffHeight);
        void FillBitmapWithSharpenFilter2D(Graphics::TBitmap *filter,Extended stromosc,Extended ffWidth,Extended ffHeight);
        void FillBitmapWithEdgeFilter2D(Graphics::TBitmap *filter,Extended level,Extended ffWidth,Extended ffHeight);
        void FlipBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,TScrollBarKind FlipAxis);
        //Do FFT
        void SwitchQuartersInTable(COMPLEX *dest,uint destWidth,uint destHeight);
        void SwitchQuartersInBitmap(Graphics::TBitmap *dest,uint destWidth,uint destHeight);
        void ConductMultiplyTable(COMPLEX *dest,uint destWidth,uint destHeight,COMPLEX *filter,uint filterWidth,uint filterHeight);
        void ConductMultipSameTbl(COMPLEX *dest,COMPLEX *filter,uint filterWidth,uint filterHeight);
        void ConductFFTBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src);
        void ConductRFTBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src);
        void FFT(COMPLEX *Y, int N);
        void FFTBitmap(COMPLEX *YR,COMPLEX *YG,COMPLEX *YB,Graphics::TBitmap *src);
        void FFTTable(COMPLEX *Y,Extended *src,uint srcWidth,uint srcHeight);
        void RFTBitmap(Graphics::TBitmap *dest,uint srcWidth,uint srcHeight,COMPLEX *YR,COMPLEX *YG,COMPLEX *YB);
        void FFTGaussianBlurBitmap(Graphics::TBitmap *dest,Graphics::TBitmap *src,Extended level);
        //Do histogramu
        void CountHistogram(THistogramType HistType,Extended *destR,Extended *destG,Extended *destB,Graphics::TBitmap *src);
        void ConductLevelsChange(Graphics::TBitmap *dest,uchar *curve,destChannel channel);
        void ConductAutoLevels(Graphics::TBitmap *dest,destChannel channel);
        void __fastcall PaintBarOnBitmap(Graphics::TBitmap* AnImage,
             const TRect& PaintRect,Graphics::TColor ForeColor,Graphics::TColor BackColor,TScrollBarKind Kind,Extended PercentDone);
        void __fastcall PaintTextOnBitmap(Graphics::TBitmap* AnImage, const TRect& PaintRect,const String &Str);
        void __fastcall PaintBitmapBackground(Graphics::TBitmap* AnImage,int CopyMode);
        void __fastcall UpdateAdvancePos();
        //Komponent z zaawansowaniem
        Extended AdvancePos;
        Extended AdvanceMax;
        //TCriticalSection *Access;
        eFilter fmode;
        Graphics::TBitmap *srcPict;
        Graphics::TBitmap *destPict;
        threadFunction function;
        Extended extParam;
        Extended extParamAng;
        int extParamInt;
        bool extParamFlag;
        Extended newWidthParam;
        Extended newHeightParam;
        Extended *FilterBuf;
        Extended *destBufR;
        Extended *destBufG;
        Extended *destBufB;
        uchar *curveParam;
};
//---------------------------------------------------------------------------
#endif
